import os
from typing import Dict, List, Optional, Union

import numpy as np

from simba.data_processors.interpolate import Interpolate
from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.checks import (check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_int,
                                check_str, check_valid_lst)
from simba.utils.enums import ConfigKey, Formats, TagNames
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (clean_superanimal_topview_filename,
                                    find_all_videos_in_project,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data,
                                    read_dlc_superanimal_h5, write_df)


class SuperAnimalTopViewImporter(PoseImporterMixin, ConfigReader):

    """
    Import SuperAnimal-TopView mouse data to SimBA

    .. note::
       For more information see the DeepLabCutModelZoo-SuperAnimal-TopViewMouse on huggingface https://huggingface.co/mwmathis/DeepLabCutModelZoo-SuperAnimal-TopViewMouse

       Trackes 27 body-parts on one or more mice recorded from zenith.

    .. image:: _static/img/superanimal_topview.png
       :width: 150
       :align: center

    :param config_path: path to SimBA project config file in Configparser format
    :param str data_folder: Path to folder containing SuperAnimal data in ``.h5`` format.
    :param List[str] id_lst: Names of animals.
    :param Optional[Dict[str, str]] interpolation_setting: Dict defining the type and method to use to perform interpolation {'type': 'animals', 'method': 'linear'}.
    :param Optional[Dict[str, Union[str, int]]] smoothing_settings: Dictionary defining the pose estimation smoothing method {'time_window': 500, 'method': 'gaussian'}.

    :references:
        .. [1] Ye, Shaokai, Anastasiia Filippova, Jessy Lauer, et al. “SuperAnimal Pretrained Pose Estimation Models for Behavioral Analysis.” Nature Communications 15, no. 1 (2024): 5165. https://doi.org/10.1038/s41467-024-48792-2.
        .. [2] mwmathis lab on huggingface - `https://huggingface.co/mwmathis/ <https://huggingface.co/mwmathis/>`_.

    :example:
    >>> importer = SuperAnimalTopViewImporter(config_path=r"C:\troubleshooting\super_animal_import\project_folder\project_config.ini", data_folder=r'C:\troubleshooting\super_animal_import\data_files', id_lst=['Animal_1'])
    >>> importer.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_folder: Union[str, os.PathLike],
                 id_lst: List[str],
                 interpolation_settings: Optional[Dict[str, str]] = None,
                 smoothing_settings: Optional[Dict[str, Union[int, str]]] = None):


        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        check_if_dir_exists(in_dir=data_folder)
        check_valid_lst(data=id_lst, source=f'{self.__class__.__name__} id_lst', valid_dtypes=(str,), min_len=1)
        if interpolation_settings is not None:
            check_if_keys_exist_in_dict(data=interpolation_settings, key=['method', 'type'], name=f'{self.__class__.__name__} interpolation_settings')
            check_str(name=f'{self.__class__.__name__} interpolation_settings type', value=interpolation_settings['type'], options=('body-parts', 'animals'))
            check_str(name=f'{self.__class__.__name__} interpolation_settings method', value=interpolation_settings['method'], options=('linear', 'quadratic', 'nearest'))
        if smoothing_settings is not None:
            check_if_keys_exist_in_dict(data=smoothing_settings, key=['method', 'time_window'], name=f'{self.__class__.__name__} smoothing_settings')
            check_str(name=f'{self.__class__.__name__} smoothing_settings method', value=smoothing_settings['method'], options=('savitzky-golay', 'gaussian'))
            check_int(name=f'{self.__class__.__name__} smoothing_settings time_window', value=smoothing_settings['time_window'], min_value=1)

        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        self.interpolation_settings, self.smoothing_settings = (interpolation_settings, smoothing_settings)
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(self.logs_path, f"data_import_log_{self.datetime}.csv")
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir, raise_error=True if len(id_lst) > 1 else False)
        self.input_data_paths = find_files_of_filetypes_in_directory(directory=self.data_folder, extensions=['.h5'], raise_error=True)
        self.data_and_videos_lk = self.link_video_paths_to_data_paths(data_paths=self.input_data_paths, video_paths=self.video_paths, raise_error=True if len(id_lst) > 1 else False, filename_cleaning_func=clean_superanimal_topview_filename)
        self.check_multi_animal_status()
        self.config.set(ConfigKey.GENERAL_SETTINGS.value, ConfigKey.ANIMAL_CNT.value, str(len(self.id_lst)))
        with open(self.config_path, "w") as f: self.config.write(f)
        f.close()
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} SuperAnimal-TopView H5 file(s)...")

    def _get_expected_column_names(self):
        self.field_names, self.bp_names = [], []
        for animal_id in self.id_lst:
            if len(self.id_lst) == 1:
                animal_field_names = [f"{animal_id}_{s}" for s in Formats.SUPERANIMAL_TOPVIEW_BP_NAMES.value]
            else:
                animal_field_names = Formats.SUPERANIMAL_TOPVIEW_BP_NAMES.value
            self.bp_names.extend((animal_field_names))
            animal_field_names = [f"{s}{suffix}" for s in animal_field_names for suffix in ("_x", "_y", "_p")]
            self.field_names.extend((animal_field_names))
        with open(self.body_parts_path, "w") as f:
            for name in self.bp_names:
                f.write(name + "\n")
        f.close()
        ConfigReader.__init__(self, config_path=self.config_path, read_video_info=False, create_logger=False)
        self.get_body_part_names()
        self.animal_bp_dict = self.create_body_part_dictionary(self.multi_animal_status, self.id_lst, len(self.id_lst), self.x_cols, self.y_cols, self.p_cols, self.clr_lst)

    def run(self):
        self._get_expected_column_names()
        for cnt, (video_name, video_data) in enumerate(self.data_and_videos_lk.items()):
            video_timer = SimbaTimer(start=True)
            self.add_spacer, self.frame_no, self.video_data, self.video_name = (2, 1, video_data, video_name)
            print(f"Processing {video_name} ({cnt+1}/{len(self.input_data_paths)})...")
            self.data_df = read_dlc_superanimal_h5(path=video_data['DATA'], col_names=self.field_names)
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(animal_bp_dict=self.animal_bp_dict, video_info=get_video_meta_data(video_data["VIDEO"]), data_df=self.data_df, video_path=video_data["VIDEO"])
                self.multianimal_identification()
            else:
                self.out_df = self.insert_multi_idx_columns(df=self.data_df.fillna(0))
            self.save_path = os.path.join(os.path.join(self.input_csv_dir, f"{self.video_name}.{self.file_type}"))
            self.out_df = self.out_df.replace([np.inf, -np.inf], np.nan).fillna(0)
            self.out_df[self.out_df < 0] = 0
            write_df(df=self.out_df, file_type=self.file_type, save_path=self.save_path, multi_idx_header=True)
            if self.interpolation_settings is not None:
                interpolator = Interpolate(config_path=self.config_path, data_path=self.save_path, type=self.interpolation_settings['type'], method=self.interpolation_settings['method'], multi_index_df_headers=True, copy_originals=False)
                interpolator.run()
            if self.smoothing_settings is not None:
                smoother = Smoothing(config_path=self.config_path, data_path=self.save_path, time_window=self.smoothing_settings['time_window'], method=self.smoothing_settings['method'], multi_index_df_headers=True, copy_originals=False)
                smoother.run()
            video_timer.stop_timer()
            stdout_success(msg=f"Video {video_name} data imported...", elapsed_time=video_timer.elapsed_time_str)
        self.timer.stop_timer()
        stdout_success(msg=f"All SuperAnimal-TopView H5 data files imported to {self.input_csv_dir} directory", elapsed_time=self.timer.elapsed_time_str)


# importer = SuperAnimalTopViewImporter(config_path=r"C:\troubleshooting\super_animal_import\project_folder\project_config.ini",
#                            data_folder=r'C:\troubleshooting\super_animal_import\data_files',
#                            id_lst=['Animal_1'])
# importer.run()


__author__ = "Simon Nilsson"

import glob
import itertools
import logging
import logging.config
import math
import os
import platform
import shutil
from configparser import ConfigParser
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty,
                                check_nvidea_gpu_available, check_str,
                                check_valid_boolean, check_valid_dataframe,
                                check_valid_lst)
from simba.utils.enums import ConfigKey, Defaults, Dtypes, Formats, Keys, Paths
from simba.utils.errors import (BodypartColumnNotFoundError, DataHeaderError,
                                DuplicationError, InvalidInputError,
                                MissingProjectConfigEntryError,
                                NoFilesFoundError, NoROIDataError,
                                NotDirectoryError, ParametersFileError,
                                PermissionError)
from simba.utils.lookups import (create_color_palettes, get_color_dict,
                                 get_emojis, get_log_config)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_core_cnt, get_all_clf_names,
                                    get_fn_ext, read_config_file, read_df,
                                    read_project_path_and_file_type, write_df)
from simba.utils.warnings import (BodypartColumnNotFoundWarning,
                                  InvalidValueWarning, NoDataFoundWarning,
                                  NoFileFoundWarning)


class ConfigReader(object):
    """
    Methods for reading SimBA configparser.Configparser project config and associated project data.

    :param configparser.Configparser config_path: path to SimBA project_config.ini
    :param bool read_video_info: if true, read the project_folder/logs/video_info.csv file.
    """

    def __init__(
        self, config_path: str, read_video_info: bool = True, create_logger: bool = True
    ):

        self.timer = SimbaTimer(start=True)
        self.config_path = config_path
        self.config = read_config_file(config_path=config_path)
        self.datetime = datetime.now().strftime("%Y%m%d%H%M%S")

        self.project_path, self.file_type = read_project_path_and_file_type(
            config=self.config
        )
        self.input_csv_dir = os.path.join(self.project_path, Paths.INPUT_CSV.value)
        self.features_dir = os.path.join(
            self.project_path, Paths.FEATURES_EXTRACTED_DIR.value
        )
        self.targets_folder = os.path.join(
            self.project_path, Paths.TARGETS_INSERTED_DIR.value
        )
        self.input_frames_dir = os.path.join(
            self.project_path, Paths.INPUT_FRAMES_DIR.value
        )
        self.machine_results_dir = os.path.join(
            self.project_path, Paths.MACHINE_RESULTS_DIR.value
        )
        self.directionality_df_dir = os.path.join(
            self.project_path, Paths.DIRECTIONALITY_DF_DIR.value
        )
        self.body_part_directionality_df_dir = os.path.join(
            self.project_path, Paths.BODY_PART_DIRECTIONALITY_DF_DIR.value
        )
        self.outlier_corrected_dir = os.path.join(
            self.project_path, Paths.OUTLIER_CORRECTED.value
        )
        self.outlier_corrected_movement_dir = os.path.join(
            self.project_path, Paths.OUTLIER_CORRECTED_MOVEMENT.value
        )
        self.heatmap_clf_location_dir = os.path.join(
            self.project_path, Paths.HEATMAP_CLF_LOCATION_DIR.value
        )
        self.heatmap_location_dir = os.path.join(
            self.project_path, Paths.HEATMAP_LOCATION_DIR.value
        )
        self.line_plot_dir = os.path.join(self.project_path, Paths.LINE_PLOT_DIR.value)
        self.probability_plot_dir = os.path.join(
            self.project_path, Paths.PROBABILITY_PLOTS_DIR.value
        )
        self.gantt_plot_dir = os.path.join(
            self.project_path, Paths.GANTT_PLOT_DIR.value
        )
        self.path_plot_dir = os.path.join(self.project_path, Paths.PATH_PLOT_DIR.value)
        self.shap_logs_path = os.path.join(self.project_path, Paths.SHAP_LOGS.value)
        self.video_info_path = os.path.join(self.project_path, Paths.VIDEO_INFO.value)
        self.cue_lights_data_dir = os.path.join(self.project_path, Paths.CUE_LIGHTS_PATH.value)
        self.frames_output_dir = os.path.join(
            self.project_path, Paths.FRAMES_OUTPUT_DIR.value
        )
        self.font = cv2.FONT_HERSHEY_COMPLEX
        self.roi_features_save_dir = os.path.join(
            self.project_path, Paths.ROI_FEATURES.value
        )
        self.configs_meta_dir = os.path.join(self.project_path, "configs")
        self.sklearn_plot_dir = os.path.join(
            self.project_path, Paths.SKLEARN_RESULTS.value
        )
        self.detailed_roi_data_dir = os.path.join(
            self.project_path, Paths.DETAILED_ROI_DATA_DIR.value
        )
        self.directing_animals_video_output_path = os.path.join(
            self.project_path, Paths.DIRECTING_BETWEEN_ANIMALS_OUTPUT_PATH.value
        )
        self.directing_body_part_animal_video_output_path = os.path.join(
            self.project_path,
            Paths.DIRECTING_BETWEEN_ANIMAL_BODY_PART_OUTPUT_PATH.value,
        )
        self.animal_cnt = self.read_config_entry(
            config=self.config,
            section=ConfigKey.GENERAL_SETTINGS.value,
            option=ConfigKey.ANIMAL_CNT.value,
            data_type=Dtypes.INT.value,
        )
        self.bodypart_direction = self.read_config_entry(
            self.config,
            section=ConfigKey.DIRECTIONALITY_SETTINGS.value,
            option=ConfigKey.BODYPART_DIRECTION_VALUE.value,
            data_type=Dtypes.STR.value,
            default_value=Dtypes.NONE.value,
        )
        self.gpu_available = check_nvidea_gpu_available()
        self.clf_cnt = self.read_config_entry(self.config, ConfigKey.SML_SETTINGS.value, ConfigKey.TARGET_CNT.value, Dtypes.INT.value)
        self.clf_names = get_all_clf_names(config=self.config, target_cnt=self.clf_cnt)
        self.feature_file_paths = glob.glob(self.features_dir + f"/*.{self.file_type}")
        self.target_file_paths = glob.glob(self.targets_folder + f"/*.{self.file_type}")
        self.input_csv_paths = glob.glob(self.input_csv_dir + f"/*.{self.file_type}")
        self.body_part_directionality_paths = glob.glob(
            self.body_part_directionality_df_dir + f"/*.{self.file_type}"
        )
        self.outlier_corrected_paths = glob.glob(
            self.outlier_corrected_dir + f"/*.{self.file_type}"
        )
        self.outlier_corrected_movement_paths = glob.glob(
            self.outlier_corrected_movement_dir + f"/*.{self.file_type}"
        )
        self.cpu_cnt, self.cpu_to_use = find_core_cnt()
        self.machine_results_paths = glob.glob(self.machine_results_dir + f"/*.{self.file_type}")
        self.logs_path = os.path.join(self.project_path, "logs")
        self.body_parts_path = os.path.join(self.project_path, Paths.BP_NAMES.value)
        check_file_exist_and_readable(file_path=self.body_parts_path)
        self.body_parts_lst = (pd.read_csv(self.body_parts_path, header=None).iloc[:, 0].to_list())
        self.body_parts_lst = [x for x in self.body_parts_lst if str(x) != "nan"]
        self.get_body_part_names()
        self.get_bp_headers()
        self.bp_col_names = self.x_cols + self.y_cols + self.p_cols
        self.pose_setting = self.read_config_entry(
            config=self.config,
            section=ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,
            option=ConfigKey.POSE_SETTING.value,
            data_type=Dtypes.STR.value,
        )
        self.roi_coordinates_path = os.path.join(
            self.logs_path, Paths.ROI_DEFINITIONS.value
        )
        self.video_dir = os.path.join(self.project_path, "videos")
        self.clf_validation_dir = os.path.join(
            self.project_path, Paths.CLF_VALIDATION_DIR.value
        )
        self.clf_data_validation_dir = os.path.join(
            self.project_path, "csv", "validation"
        )
        self.annotated_frm_dir = os.path.join(
            self.project_path, Paths.ANNOTATED_FRAMES_DIR.value
        )
        self.single_validation_video_save_dir = os.path.join(
            self.project_path, Paths.SINGLE_CLF_VALIDATION.value
        )
        self.data_table_path = os.path.join(self.project_path, Paths.DATA_TABLE.value)
        self.check_multi_animal_status()
        self.multiprocess_chunksize = Defaults.CHUNK_SIZE.value
        self.maxtasksperchild = Defaults.MAX_TASK_PER_CHILD.value
        self.clr_lst = create_color_palettes(self.animal_cnt, int(len(self.x_cols)) + 1)
        self.animal_bp_dict = self.create_body_part_dictionary(
            self.multi_animal_status,
            self.multi_animal_id_list,
            self.animal_cnt,
            self.x_cols,
            self.y_cols,
            self.p_cols,
            self.clr_lst,
        )
        self.project_bps = list(set([x[:-2] for x in self.bp_headers]))
        self.platform = platform.system()
        self.color_dict = get_color_dict()
        if create_logger:
            self.create_logger()
        self.emojis = get_emojis()
        if read_video_info:
            self.video_info_df = self.read_video_info_csv(file_path=self.video_info_path)


        self.min_draw_display_ratio_h = self.read_config_entry(config=self.config, section=ConfigKey.DISPLAY_SETTINGS.value, option=ConfigKey.MIN_ROI_DISPLAY_HEIGHT.value, default_value=0.20, data_type=Dtypes.FLOAT.value)
        self.min_draw_display_ratio_w = self.read_config_entry(config=self.config, section=ConfigKey.DISPLAY_SETTINGS.value, option=ConfigKey.MIN_ROI_DISPLAY_WIDTH.value, default_value=0.20, data_type=Dtypes.FLOAT.value)
        self.max_draw_display_ratio_h = self.read_config_entry(config=self.config, section=ConfigKey.DISPLAY_SETTINGS.value, option=ConfigKey.MAX_ROI_DISPLAY_HEIGHT.value, default_value=0.75, data_type=Dtypes.FLOAT.value)
        self.max_draw_display_ratio_w = self.read_config_entry(config=self.config, section=ConfigKey.DISPLAY_SETTINGS.value, option=ConfigKey.MAX_ROI_DISPLAY_WIDTH.value, default_value=0.50, data_type=Dtypes.FLOAT.value)

    def read_roi_data(self) -> None:
        """
        Method to read in ROI definitions from SimBA project
        """

        if not os.path.isfile(self.roi_coordinates_path):
            raise NoROIDataError(msg="SIMBA ERROR: No ROI definitions were found in your SimBA project. Please draw some ROIs before analyzing your ROI data", source=self.__class__.__name__)
        else:
            self.rectangles_df = pd.read_hdf(self.roi_coordinates_path, key=Keys.ROI_RECTANGLES.value)
            if ("Center_X" in self.rectangles_df.columns) and (self.rectangles_df["Center_X"].isnull().values.any()):
                for idx, row in self.rectangles_df.iterrows():
                    self.rectangles_df.loc[idx]["Center_X"] = row["Tags"]["Center tag"][0]
                    self.rectangles_df.loc[idx]["Center_Y"] = row["Tags"]["Center tag"][1]
            self.circles_df = pd.read_hdf(self.roi_coordinates_path, key=Keys.ROI_CIRCLES.value).dropna(how="any")
            self.polygon_df = pd.read_hdf(self.roi_coordinates_path, key=Keys.ROI_POLYGONS.value)
            if "Center_XCenter_Y" in self.polygon_df.columns:
                self.polygon_df = self.polygon_df.drop(["Center_XCenter_Y"], axis=1)
            self.polygon_df = self.polygon_df.dropna(how="any")
            self.shape_names = list(itertools.chain(self.rectangles_df["Name"].unique(), self.polygon_df["Name"].unique(), self.circles_df["Name"].unique()))
            self.roi_names = deepcopy(self.shape_names)
            self.roi_dict = {Keys.ROI_RECTANGLES.value: self.rectangles_df, Keys.ROI_CIRCLES.value: self.circles_df, Keys.ROI_POLYGONS.value: self.polygon_df}
            self.roi_types_names_lst = set()

            #self.roi_dict[Keys.ROI_CIRCLES.value]['Video'] = self.roi_dict[Keys.ROI_CIRCLES.value]['Video'].replace('Trial     1_dSLR1_sample_A1_na', '501_MA142_Gi_Saline_0513')

            for idx, r in self.roi_dict[Keys.ROI_RECTANGLES.value].iterrows():
                self.roi_types_names_lst.add(f'Rectangle: {r["Name"]}')
            for idx, r in self.roi_dict[Keys.ROI_CIRCLES.value].iterrows():
                self.roi_types_names_lst.add(f'Circle: {r["Name"]}')
            for idx, r in self.roi_dict[Keys.ROI_POLYGONS.value].iterrows():
                self.roi_types_names_lst.add(f'Polygon: {r["Name"]}')
            self.roi_types_names_lst = list(self.roi_types_names_lst)
            for shape_type, shape_data in self.roi_dict.items():
                if shape_type == Keys.ROI_CIRCLES.value:
                    self.roi_dict[Keys.ROI_CIRCLES.value]["Center_X"] = self.roi_dict[Keys.ROI_CIRCLES.value]["centerX"]
                    self.roi_dict[Keys.ROI_CIRCLES.value]["Center_Y"] = self.roi_dict[Keys.ROI_CIRCLES.value]["centerY"]
                elif shape_type == Keys.ROI_RECTANGLES.value:
                    self.roi_dict[Keys.ROI_RECTANGLES.value]["Center_X"] = self.roi_dict[Keys.ROI_RECTANGLES.value]["Bottom_right_X"] - (self.roi_dict[Keys.ROI_RECTANGLES.value]["width"] / 2)
                    self.roi_dict[Keys.ROI_RECTANGLES.value]["Center_Y"] = self.roi_dict[Keys.ROI_RECTANGLES.value]["Bottom_right_Y"] - (self.roi_dict[Keys.ROI_RECTANGLES.value]["height"] / 2)
                elif shape_type == Keys.ROI_POLYGONS.value:
                    try:
                        self.roi_dict[Keys.ROI_POLYGONS.value]["Center_X"] = (self.roi_dict[Keys.ROI_POLYGONS.value]["Center_X"])
                        self.roi_dict[Keys.ROI_POLYGONS.value]["Center_Y"] = (self.roi_dict[Keys.ROI_POLYGONS.value]["Center_Y"])
                    except KeyError:
                        pass
            self.video_names_w_rois = set(list(self.rectangles_df["Video"]) + list(self.circles_df["Video"]) + list(self.polygon_df["Video"]))

    def get_all_clf_names(self) -> List[str]:
        """
        Helper to return all classifier names in SimBA project

        :return List[str]:
        """

        model_names = []
        for i in range(self.clf_cnt):
            entry_name = "target_name_{}".format(str(i + 1))
            model_names.append(
                self.read_config_entry(
                    self.config,
                    ConfigKey.SML_SETTINGS.value,
                    entry_name,
                    data_type=Dtypes.STR.value,
                )
            )
        return model_names

    def insert_column_headers_for_outlier_correction(
        self, data_df: pd.DataFrame, new_headers: List[str], filepath: str
    ) -> pd.DataFrame:
        """
        Helper to insert new column headers onto a dataframe.

        :param pd.DataFrame data_df:  Dataframe where headers to to-bo replaced.
        :param List[str] new_headers:  Names of new headers.
        :param str filepath:  Path to where ``data_df`` is stored on disk
        :returns pd.DataFrame: Dataframe with new headers
        :raises DataHeaderWarning: If new headers are fewer/more than columns in dataframe

        :example:
        >>> df = pd.DataFrame(data=[[1, 2, 3], [1, 2, 3]], columns=['Feature_1', 'Feature_2', 'Feature_3'])
        >>> ConfigReader.insert_column_headers_for_outlier_correction(data_df=df, new_headers=['Feature_4', 'Feature_5', 'Feature_6'], filepath='test/my_test_file.csv')
        """

        if len(new_headers) != len(data_df.columns):
            difference = int(len(data_df.columns) - len(new_headers))
            bp_missing = int(abs(difference) / 3)
            if difference < 0:
                raise DataHeaderError(
                    msg=f"SIMBA ERROR: SimBA expects {len(new_headers)} columns of data inside the files within project_folder/csv/input_csv directory. However, within file {filepath} file, SimBA found {len(data_df.columns)} columns. Thus, there is {abs(difference)} missing data columns in the imported data, which may represent {int(bp_missing)} bodyparts if each body-part has an x, y and p value. Either revise the SimBA project pose-configuration with {bp_missing} less body-part, or include {bp_missing} more body-part in the imported data",
                    source=self.__class__.__name__,
                )
            else:
                raise DataHeaderError(
                    msg=f"SIMBA ERROR: SimBA expects {len(new_headers)} columns of data inside the files within project_folder/csv/input_csv directory. However, within file {filepath} file, SimBA found {len(data_df.columns)} columns. Thus, there is {abs(difference)} more data columns in the imported data than anticipated, which may represent {int(bp_missing)} bodyparts if each body-part has an x, y and p value. Either revise the SimBA project pose-configuration with {bp_missing} more body-part, or include {bp_missing} less body-part in the imported data",
                    source=self.__class__.__name__,
                )
        else:
            data_df.columns = new_headers
            return data_df

    def get_number_of_header_columns_in_df(self, df: pd.DataFrame) -> int:
        """
        Helper to find the count of non-numerical rows at the top of a dataframe.

        :param pd.DataFrame data_df: Dataframe to find the count non-numerical header rows in.
        :returns int
        :raises DataHeaderError: All rows are non-numerical.

        :example:
        >>> ConfigReader.get_number_of_header_columns_in_df(df=pd.DataFrame(data=[[1, 2, 3], [1, 2, 3]]))
        >>> 0
        >>> ConfigReader.get_number_of_header_columns_in_df(df=pd.DataFrame(data=[['Head_1', 'Body_2', 'Tail_3'], ['Some_nonsense', 'A_mistake', 'Maybe_multi_headers?'], [11, 99, 109], [122, 43, 2091]]))
        >>> 2
        """

        for i in range(len(df)):
            try:
                _ = df.iloc[i:].apply(pd.to_numeric).reset_index(drop=True)
                return i
            except ValueError:
                pass
        raise DataHeaderError(
            msg="Could find the count of header columns in dataframe. All values appear to be non-numeric",
            source=self.__class__.__name__,
        )

    def find_video_of_file(
        self,
        video_dir: Union[str, os.PathLike],
        filename: str,
        raise_error: bool = False,
    ) -> Union[str, os.PathLike]:
        """
        Helper to find the video file representing a known data file basename.

        :param Union[str, os.PathLike] video_dir: Directory holding putative video file.
        :param str filename: Data file name, e.g., ``Video_1``.
        :param bool raise_error: If True, raise error if no video can be found.
        :return Union[str, os.PathLike]: Path to video file.


        :example:
        >>> config_reader = ConfigReader(config_path='My_SimBA_Config')
        >>> config_reader.find_video_of_file(video_dir=config_reader.video_dir, filename='Video1')
        >>> '/project_folder/videos/Video1.mp4'
        """
        try:
            all_files_in_video_folder = [
                f for f in next(os.walk(video_dir))[2] if not f[0] == "."
            ]
        except StopIteration:
            raise NoFilesFoundError(
                msg=f"No files found in the {video_dir} directory",
                source=self.__class__.__name__,
            )
        all_files_in_video_folder = [
            os.path.join(video_dir, x) for x in all_files_in_video_folder
        ]
        return_path = None
        for file_path in all_files_in_video_folder:
            _, video_filename, ext = get_fn_ext(file_path)
            if (video_filename == filename) and (
                (ext.lower() == ".mp4") or (ext.lower() == ".avi")
            ):
                return_path = file_path

        if return_path is None and not raise_error:
            NoFileFoundWarning(
                msg=f"SimBA could not find a video file representing {filename} in the project video directory",
                source=self.__class__.__name__,
            )
        if return_path is None and raise_error:
            raise NoFilesFoundError(
                msg=f"SimBA could not find a video file representing {filename} in the project video directory",
                source=self.__class__.__name__,
            )
        return return_path

    def add_missing_ROI_cols(self, shape_df: pd.DataFrame) -> pd.DataFrame:
        """
        Helper to add missing ROI definitions (``Color BGR``, ``Thickness``, ``Color name``) in ROI info
        dataframes created by the first version of the SimBA ROI user-interface but analyzed using newer versions of SimBA.


        :param pd.DataFrame shape_df: Dataframe holding ROI definitions.
        :return pd.DataFrame with ``Color BGR``, ``Thickness``, ``Color name`` fields
        """

        if not "Color BGR" in shape_df.columns:
            shape_df["Color BGR"] = [(255, 255, 255)] * len(shape_df)
        if not "Thickness" in shape_df.columns:
            shape_df["Thickness"] = [5] * len(shape_df)
        if not "Color name" in shape_df.columns:
            shape_df["Color name"] = "White"

        return shape_df

    def remove_a_folder(
        self, folder_dir: str, raise_error: Optional[bool] = False
    ) -> None:
        """
        Helper to remove single directory.

        :param folder_dir: Directory to remove.
        :param bool raise_error: If True, raise ``NotDirectoryError`` error of folder does not exist.
        :raises NotDirectoryError: If ``raise_error`` and directory does not exist.

        :example:
        >>> self.remove_a_folder(folder_dir'gerbil/gerbil_data/featurized_data/temp')
        """
        shutil.rmtree(folder_dir, ignore_errors=True)

    def remove_multiple_folders(
        self, folders: List[os.PathLike], raise_error: Optional[bool] = False
    ) -> None:
        """
        Helper to remove multiple directories.

        :param folders List[os.PathLike]: List of directory paths.
        :param bool raise_error: If True, raise ``NotDirectoryError`` error of folder does not exist. if False, then pass. Default False.
        :raises NotDirectoryError: If ``raise_error`` and directory does not exist.

        :example:
        >>> self.remove_multiple_folders(folders= ['gerbil/gerbil_data/featurized_data/temp'])
        """
        folders = [x for x in folders if x is not None]
        for folder_path in folders:
            if raise_error and not os.path.isdir(folder_path):
                raise NotDirectoryError(
                    msg=f"Cannot delete directory {folder_path}. The directory does not exist.",
                    source=self.__class__.__name__,
                )
            if os.path.isdir(folder_path):
                shutil.rmtree(folder_path, ignore_errors=True)
            else:
                pass

    def find_animal_name_from_body_part_name(self, bp_name: str, bp_dict: dict) -> str:
        """
        Given body-part name and animal body-part dict, returns the animal name

        :param str bp_name: Name of the body-part. E.g., ``Ear_1``.
        :param dict bp_dict: Nested dict holding animal names as keys and body-part names as and coordinates as values. Created by :meth:`simba.mixins.config_reader.ConfigReader.create_body_part_dictionary`
        :returns str

        :example:
        >>> config_reader = ConfigReader(config_path='tests/data/test_projects/two_c57/project_folder/project_config.ini')
        >>> ConfigReader.find_animal_name_from_body_part_name(bp_name='Ear_1', bp_dict=config_reader.animal_bp_dict)
        >>> 'simon'
        """

        for animal_name, animal_bps in bp_dict.items():
            if bp_name in [x[:-2] for x in animal_bps["X_bps"]]:
                return animal_name

    def create_body_part_dictionary(
        self,
        multi_animal_status: bool,
        animal_id_lst: list,
        animal_cnt: int,
        x_cols: List[str],
        y_cols: List[str],
        p_cols: Optional[List[str]] = None,
        colors: Optional[List[List[Tuple[int, int, int]]]] = None,
    ) -> Dict[str, Union[List[str], List[Tuple]]]:
        """
        Helper to create dict of dict lookup of body-parts where the keys are animal names, and
        values are the body-part names.

        :param bool multi_animal_status: If True, it is a multi-animal SimBA project.
        :param List[str] multi_animal_id_lst: Animal names. Eg., ['Simon, 'JJ']. Note: If a single animal project, this will be overridden and set to `Animal_1`.
        :param int animal_cnt: Number of animals in the SimBA project.
        :param List[str] x_cols: column names for body-part coordinates on x-axis. Returned by  :meth:`simba.mixins.config_reader.ConfigReader.get_body_part_names`
        :param List[str] y_cols: column names for body-part coordinates on y-axis. Returned by  :meth:`simba.mixins.config_reader.ConfigReader.get_body_part_names`
        :param List[str] p_cols: column names for body-part pose-estimation probability values. Returned by  :meth:`simba.mixins.config_reader.ConfigReader.get_body_part_names`
        :param Optional[List[List[Tuple[int, int, int]]]] colors: Optional bgr colors to associate with the body-parts. Returned by :meth:`simba.utils.data.create_color_palettes`.
        :returns dict

        :example:
        >>> ConfigReader.create_body_part_dictionary(multi_animal_status=True, animal_id_lst=['simon',])
        >>> {'simon': {'X_bps': ['Nose_1_x', 'Ear_left_1_x', 'Ear_right_1_x', 'Center_1_x', 'Lat_left_1_x', 'Lat_right_1_x', 'Tail_base_1_x', 'Tail_end_1_x'], 'Y_bps': ['Nose_1_y', 'Ear_left_1_y', 'Ear_right_1_y', 'Center_1_y', 'Lat_left_1_y', 'Lat_right_1_y', 'Tail_base_1_y', 'Tail_end_1_y'], 'colors': [[255.0, 0.0, 255.0], [223.125, 31.875, 255.0], [191.25, 63.75, 255.0], [159.375, 95.625, 255.0], [127.5, 127.5, 255.0], [95.625, 159.375, 255.0], [63.75, 191.25, 255.0], [31.875, 223.125, 255.0], [0.0, 255.0, 255.0]], 'P_bps': ['Nose_1_p', 'Ear_left_1_p', 'Ear_right_1_p', 'Center_1_p', 'Lat_left_1_p', 'Lat_right_1_p', 'Tail_base_1_p', 'Tail_end_1_p']}, 'jj': {'X_bps': ['Nose_2_x', 'Ear_left_2_x', 'Ear_right_2_x', 'Center_2_x', 'Lat_left_2_x', 'Lat_right_2_x', 'Tail_base_2_x', 'Tail_end_2_x'], 'Y_bps': ['Nose_2_y', 'Ear_left_2_y', 'Ear_right_2_y', 'Center_2_y', 'Lat_left_2_y', 'Lat_right_2_y', 'Tail_base_2_y', 'Tail_end_2_y'], 'colors': [[102.0, 127.5, 0.0], [102.0, 143.4375, 31.875], [102.0, 159.375, 63.75], [102.0, 175.3125, 95.625], [102.0, 191.25, 127.5], [102.0, 207.1875, 159.375], [102.0, 223.125, 191.25], [102.0, 239.0625, 223.125], [102.0, 255.0, 255.0]], 'P_bps': ['Nose_2_p', 'Ear_left_2_p', 'Ear_right_2_p', 'Center_2_p', 'Lat_left_2_p', 'Lat_right_2_p', 'Tail_base_2_p', 'Tail_end_2_p']}}
        """

        animal_bp_dict = {}

        if multi_animal_status:
            for animal in range(animal_cnt):
                animal_bp_dict[animal_id_lst[animal]] = {}
                animal_bp_dict[animal_id_lst[animal]]["X_bps"] = [
                    i for i in x_cols if animal_id_lst[animal] in i
                ]
                animal_bp_dict[animal_id_lst[animal]]["Y_bps"] = [
                    i for i in y_cols if animal_id_lst[animal] in i
                ]
                if colors:
                    animal_bp_dict[animal_id_lst[animal]]["colors"] = colors[animal]
                if p_cols:
                    animal_bp_dict[animal_id_lst[animal]]["P_bps"] = [
                        i for i in p_cols if animal_id_lst[animal] in i
                    ]
                if not animal_bp_dict[animal_id_lst[animal]]["X_bps"]:
                    multi_animal_status = False
                    break

        if not multi_animal_status:
            if animal_cnt > 1:
                for animal in range(animal_cnt):
                    currAnimalName = f"Animal_{str(animal + 1)}"
                    search_string_x = f"_{str(animal + 1)}_x"
                    search_string_y = f"_{str(animal + 1)}_y"
                    search_string_p = f"_{str(animal + 1)}_p"
                    animal_bp_dict[currAnimalName] = {}
                    animal_bp_dict[currAnimalName]["X_bps"] = [
                        i for i in x_cols if i.endswith(search_string_x)
                    ]
                    animal_bp_dict[currAnimalName]["Y_bps"] = [
                        i for i in y_cols if i.endswith(search_string_y)
                    ]
                    if colors:
                        animal_bp_dict[currAnimalName]["colors"] = colors[animal]
                    if p_cols:
                        animal_bp_dict[currAnimalName]["P_bps"] = [
                            i for i in p_cols if i.endswith(search_string_p)
                        ]
                if animal_id_lst[0] != "":
                    for animal in range(len(animal_id_lst)):
                        currAnimalName = f"Animal_{str(animal + 1)}"
                        animal_bp_dict[animal_id_lst[animal]] = animal_bp_dict.pop(
                            currAnimalName
                        )

            else:
                animal_bp_dict["Animal_1"] = {}
                animal_bp_dict["Animal_1"]["X_bps"] = [i for i in x_cols]
                animal_bp_dict["Animal_1"]["Y_bps"] = [i for i in y_cols]
                if colors:
                    animal_bp_dict["Animal_1"]["colors"] = colors[0]
                if p_cols:
                    animal_bp_dict["Animal_1"]["P_bps"] = [i for i in p_cols]
        return animal_bp_dict

    def get_body_part_names(self):
        """
        Helper to extract pose-estimation data field names (x, y, p)

        :example:
        >>> config_reader = ConfigReader(config_path='test/project_config.csv')
        >>> config_reader.get_body_part_names()
        """

        self.x_cols, self.y_cols, self.p_cols = [], [], []
        for bp in self.body_parts_lst:
            self.x_cols.append(f"{bp}_x")
            self.y_cols.append(f"{bp}_y")
            self.p_cols.append(f"{bp}_p")

    def drop_bp_cords(
        self, df: pd.DataFrame, raise_error: bool = False
    ) -> pd.DataFrame:
        """
        Helper to remove pose-estimation fields from dataframe.

        :param pd.DataFrame df: pandas dataframe containing pose-estimation fields (body-part x, y, p fields)
        :param bool raise_error: If True, raise error if body-parts cant be found. Else, print warning
        :return pd.DataFrame: ``df`` without pose-estimation fields

        :example:
        >>> config_reader = ConfigReader(config_path='test/project_folder/project_config.csv')
        >>> df = read_df(config_reader.machine_results_paths[0], file_type='csv')
        >>> df = config_reader.drop_bp_cords(df=df)
        """
        missing_body_part_fields = list(set(self.bp_col_names) - set(list(df.columns)))
        if len(missing_body_part_fields) > 0 and not raise_error:
            BodypartColumnNotFoundWarning(
                msg=f"SimBA could not drop body-part coordinates, some body-part names are missing in dataframe. SimBA expected the following body-parts, that could not be found inside the file: {missing_body_part_fields}",
                source=self.__class__.__name__,
            )
            return df.drop(self.bp_col_names, axis=1, errors="ignore")
        elif len(missing_body_part_fields) > 0 and raise_error:
            raise BodypartColumnNotFoundError(
                msg=f"SimBA could not drop body-part coordinates, some body-part names are missing in dataframe. SimBA expected the following body-parts, that could not be found inside the file: {missing_body_part_fields}",
                source=self.__class__.__name__,
            )
        else:
            return df.drop(self.bp_col_names, axis=1)

    def get_bp_headers(self) -> None:
        """
        Helper to create ordered list of all column header fields for SimBA project dataframes.

        >>> config_reader = ConfigReader(config_path='test/project_folder/project_config.ini')
        >>> config_reader.get_bp_headers()
        """

        self.bp_headers = []
        for bp in self.body_parts_lst:
            c1, c2, c3 = (f"{bp}_x", f"{bp}_y", f"{bp}_p")
            self.bp_headers.extend((c1, c2, c3))

    def read_config_entry(
        self,
        config: ConfigParser,
        section: str,
        option: str,
        data_type: Literal["str", "int", "float", "folder_path"],
        default_value: Optional[Any] = None,
        options: Optional[List[Any]] = None,
    ) -> Union[str, int, float]:
        """
        Helper to read entry from a configparser.ConfigParser object

        :param ConfigParser config: Parsed SimBA project config
        :param str section: Project config section name
        :param str option: Project config option name
        :param str data_type: Type of data. E.g., ``str``, ``int``, ``float``, ``folder_path``.
        :param Optional[Any] default_value: If entry cannot be found, then default to this value.
        :param Optional[List[Any]] options: If not ``None``, then list of viable entries.
        :raise InvalidInputError: If returned value is not in ``options``.
        :raise MissingProjectConfigEntryError: If no entry is found and no ``default_value`` is provided.

        :return Union[str, float, int, os.Pathlike]

        :example:
        >>> config = ConfigReader(config_path='tests/data/test_projects/two_c57/project_folder/project_config.ini')
        >>> config.read_config_entry(config=self.config, section='Multi animal IDs', option='id_list', data_type='str')
        >>> 'simon,jj'
        """

        try:
            if config.has_option(section, option):
                if data_type == Dtypes.FLOAT.value:
                    value = config.getfloat(section, option)
                elif data_type == Dtypes.INT.value:
                    value = config.getint(section, option)
                elif data_type == Dtypes.STR.value:
                    value = config.get(section, option).strip()
                elif data_type == Dtypes.FOLDER.value:
                    value = config.get(section, option).strip()
                    if not os.path.isdir(value):
                        raise NotDirectoryError(
                            msg=f"The SimBA config file includes paths to a folder ({value}) that does not exist.",
                            source=self.__class__.__name__,
                        )
                if options != None:
                    if value not in options:
                        raise InvalidInputError(
                            msg=f"{option} is set to {str(value)} in SimBA, but this is not among the valid options: ({options})",
                            source=self.__class__.__name__,
                        )
                    else:
                        return value
                return value
            elif default_value != None:
                return default_value
            else:
                raise MissingProjectConfigEntryError(
                    msg=f"SimBA could not find an entry for option {option} under section {section} in the project_config.ini. Please specify the settings in the settings menu.",
                    source=self.__class__.__name__,
                )
        except ValueError as e:
            if default_value != None:
                return default_value
            else:
                raise MissingProjectConfigEntryError(
                    msg=f"SimBA could not find an entry for option {option} under section {section} in the project_config.ini. Please specify the settings in the settings menu.",
                    source=self.__class__.__name__,
                )

    def read_video_info_csv(self, file_path: Union[str, os.PathLike]) -> pd.DataFrame:
        """
        Helper to read the project_folder/logs/video_info.csv of the SimBA project in as a pd.DataFrame

        :param Union[str, os.PathLike] file_path: Path to the project_folder/logs/video_info.csv file.
        :return: Dataframe representation of the file.
        :rtype: pd.DataFrame
        """

        EXPECTED_COLS = ["Video", "fps", "Resolution_width", "Resolution_height", "Distance_in_mm", "pixels/mm"]
        EXPECTED_FLOAT_COLS = ["fps", "Resolution_width", "Resolution_height", "Distance_in_mm", "pixels/mm"]

        if not os.path.isfile(file_path):
            raise NoFilesFoundError(msg=f"Could not find the video_info.csv table in your SimBA project. Create it using the [Video parameters] tab. SimBA expects the file at location {file_path}. See SimBA documentation for more info: https://t.ly/OtY79", source=ConfigReader.read_video_info_csv.__name__)
        info_df = pd.read_csv(file_path)
        for c in EXPECTED_COLS:
            if c not in info_df.columns:
                raise ParametersFileError(msg=f'The file {file_path} does not not have an anticipated header ({c}). Please re-create the file and make sure each video has a {c} header column name', source=ConfigReader.read_video_info_csv.__name__)
        info_df["Video"] = info_df["Video"].astype(str)
        for c in EXPECTED_FLOAT_COLS:
            col_vals = list(info_df[c])
            validity = check_valid_lst(data=col_vals, source='', valid_dtypes=Formats.NUMERIC_DTYPES.value, raise_error=False)
            if not validity:
                raise ParametersFileError(msg=f'One or more values in the {c} column of the {file_path} file could not be interpreted as a numeric value. Please check or re-create the file and make sure the entries in the {c} column are all numeric.', source=ConfigReader.read_video_info_csv.__name__)
            else:
                pass
        if info_df["fps"].min() <= 1:
            videos_w_low_fps = ', '.join(list(info_df[info_df['fps'] <= 1]['Video']))
            InvalidValueWarning(msg=f"Video(s) in your SimBA project have an FPS of 1 or less. This includes video(s) {videos_w_low_fps}. It is recommended to use videos with more than one frame per second. If inaccurate, correct the FPS values inside the {file_path} file", source=ConfigReader.read_video_info_csv.__name__)
        if info_df["pixels/mm"].min() == 0:
            videos_w_low_conversion_factor = ', '.join(list(info_df[info_df['pixels/mm'] == 0]['Video']))
            InvalidValueWarning(msg=f"Video(s) in your SimBA project have an pixel/mm conversion factor of 0. This includes video(s) {videos_w_low_conversion_factor}. Correct the pixel/mm conversion factor values inside the {file_path} file", source=ConfigReader.read_video_info_csv.__name__)

        return info_df

    def read_video_info(self,
                        video_name: str,
                        video_info_df_path: Optional[Union[str, os.PathLike]] = None,
                        raise_error: Optional[bool] = True) -> Union[Tuple[pd.DataFrame, float, float], Tuple[None, None, None]]:

        """
        Helper to read the meta-data (pixels per mm, resolution, fps) from the video_info.csv for a single input file.

        :param str video_name: The name of the video without extension to get the metadata for
        :param Optional[bool] raise_error: If True, raise error if video info for the video name cannot be found. Default: True.
        :param Optional[Union[str, os.PathLike]] video_info_df_path: Optional path to the video_info.csv on disk. If None, uses path from self.

        :returns: Tuple representing all video info, pixels per mm, and fps. If raise_error is False, returns a tuple of Nones
        :rtype: Union[Tuple[pd.DataFrame, float, float], Tuple[None, None, None]]
        """

        check_str(name=f'{ConfigReader.read_video_info.__name__} video_name', value=video_name, allow_blank=False)
        check_valid_boolean(value=[raise_error], source=f'{ConfigReader.read_video_info.__name__} raise_error')
        if video_info_df_path is not None:
            check_file_exist_and_readable(file_path=video_info_df_path)
            video_info_df = self.read_video_info_csv(file_path=video_info_df_path)
        else:
            video_info_df = deepcopy(self.video_info_df)
            video_info_df_path = deepcopy(self.video_info_path)
        check_valid_dataframe(df=video_info_df, source='', required_fields=["Video", "pixels/mm", "fps"])
        video_settings = video_info_df.loc[video_info_df["Video"] == video_name]
        if len(video_settings) > 1:
            raise DuplicationError(msg=f"SimBA found multiple rows in the {video_info_df_path} named {video_name}. Please make sure that each video name is represented ONCE in the {video_info_df_path} file", source=ConfigReader.read_video_info.__name__)
        elif len(video_settings) < 1:
            if raise_error:
                raise ParametersFileError(msg=f"SimBA could not find {video_name} in the video_info.csv file. Make sure all videos analyzed are represented in the {video_info_df_path} file.", source=ConfigReader.read_video_info.__name__)
            else:
                return (None, None, None)
        else:
            px_per_mm = video_settings["pixels/mm"].values[0]
            fps = video_settings["fps"].values[0]
            if math.isnan(px_per_mm):
                raise ParametersFileError(msg=f'Pixels per millimeter for video {video_name} in the {video_info_df_path} file is not a valid number.')
            if math.isnan(fps):
                raise ParametersFileError(msg=f'The FPS for video {video_name} in the {video_info_df_path} file is not a valid number.')
            check_float(name='px_per_mm', value=px_per_mm)
            check_float(name='fps', value=fps)
            px_per_mm, fps = float(px_per_mm), float(fps)
            if px_per_mm <= 0:
                InvalidValueWarning(msg=f"Video {video_name} has a pixel per millimeter conversion factor of 0 or less. Correct the pixel/mm conversion factor values inside the {video_info_df_path} file", source=ConfigReader.read_video_info.__name__)
            if fps <= 1:
                InvalidValueWarning(msg=f"Video {video_name} an FPS of 1 or less.  It is recommended to use videos with more than one frame per second. If inaccurate, correct the FPS values inside the {video_info_df_path} file", source=ConfigReader.read_video_info.__name__)
            return video_settings, px_per_mm, fps

    def check_multi_animal_status(self) -> None:
        """
        Helper to check if the project is a multi-animal SimBA project.

        """
        multi_animal_id_lst = []
        if not self.config.has_section("Multi animal IDs"):
            for animal in range(self.animal_cnt):
                multi_animal_id_lst.append("Animal_" + str(animal + 1))
            multi_animal_status = False

        else:
            multi_animal_id_str = self.read_config_entry(
                config=self.config,
                section="Multi animal IDs",
                option="id_list",
                data_type="str",
            )
            multi_animal_id_lst = [x.lstrip() for x in multi_animal_id_str.split(",")]
            multi_animal_id_lst = [x for x in multi_animal_id_lst if x != "None"]
            if (self.animal_cnt > 1) and (len(multi_animal_id_lst) > 1):
                multi_animal_status = True
            else:
                for animal in range(self.animal_cnt):
                    multi_animal_id_lst.append("Animal_{}".format(str(animal + 1)))
                multi_animal_status = False

        self.multi_animal_status = multi_animal_status
        self.multi_animal_id_list = multi_animal_id_lst[: self.animal_cnt]

    def remove_roi_features(self, data_dir: Union[str, os.PathLike]) -> None:
        """
        Helper to remove ROI-based features from datasets within a directory. The identified ROI-based fields are move to the
        ``project_folder/logs/ROI_data_{datetime}`` directory.

        .. note::
           ROI-based features are identified based on the combined criteria of (i) The prefix of the field is a named ROI in
           the ``project_folder/logs/ROI_definitions.h5`` file, and (ii) the suffix of the field is contained in the ['in zone',
           'n zone_cumulative_time', 'in zone_cumulative_percent', 'distance', 'facing']

        :param Union[str, os.PathLike] data_dir: directory with data to remove ROi features from.

        :example:
        >>> self.remove_roi_features('/project_folder/csv/features_extracted')

        """
        ROI_COL_SUFFIXES = [
            "in zone",
            "n zone_cumulative_time",
            "in zone_cumulative_percent",
            "distance",
            "facing",
        ]

        check_if_dir_exists(in_dir=data_dir)
        timer = SimbaTimer(start=True)
        filepaths = glob.glob(data_dir + f"/*.{self.file_type}")
        roi_dir = os.path.join(
            self.project_path,
            "logs",
            f'ROI_data_{datetime.now().strftime("%Y%m%d%H%M%S")}',
        )
        check_if_filepath_list_is_empty(
            filepaths=filepaths,
            error_msg=f"No .{self.file_type} files found in {data_dir}.",
        )
        for file_cnt, file_path in enumerate(filepaths):
            _, video_name, _ = get_fn_ext(filepath=file_path)
            roi_cols = set()
            df = read_df(file_path=file_path, file_type=self.file_type)
            for shape_name in self.shape_names:
                roi_cols.update([x for x in df.columns if shape_name in x])
            for suffix in ROI_COL_SUFFIXES:
                roi_cols.update([x for x in roi_cols if x.endswith(suffix)])
            if len(roi_cols) == 0:
                NoDataFoundWarning(
                    msg=f"NO ROI data found in {file_path}",
                    source=self.__class__.__name__,
                )
            else:
                roi_df = df[list(roi_cols)]
                df = df.drop(list(roi_cols), axis=1)
                roi_save_path = os.path.join(roi_dir, video_name + "." + self.file_type)
                if not os.path.exists(roi_dir):
                    os.makedirs(roi_dir)
                write_df(df=roi_df, file_type=self.file_type, save_path=roi_save_path)
                write_df(df=df, file_type=self.file_type, save_path=file_path)
                print(f"ROI features (N={len(roi_cols)}) removed from {video_name}...")
        timer.stop_timer()
        stdout_success(
            msg=f"ROI features removed from {len(filepaths)} files in the {data_dir} directory. The ROI features has been moved to the {roi_dir} directory",
            elapsed_time=timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def create_log_msg_from_init_args(self, locals: dict):

        def has_methods_or_classes(obj):
            return any(callable(attr) for attr in dir(obj))

        def remove_keys_with_methods_or_classes(dictionary):
            return {
                key: value
                for key, value in dictionary.items()
                if not has_methods_or_classes(value)
            }

        msg = ""
        locals.pop("self", None)
        locals.pop("__class__", None)
        locals = remove_keys_with_methods_or_classes(locals)

        for cnt, (k, v) in enumerate(locals.items()):
            msg += f"{k}: {v}"
            if cnt != len(list(locals.keys())) - 1:
                msg += ", "
        return msg

    def create_logger(self) -> None:
        try:
            if self.__class__ not in logging.Logger.manager.loggerDict.keys():
                log_config = get_log_config()
                log_config["handlers"]["file_handler"]["filename"] = os.path.join(self.logs_path, "project_log.log")
                logging.config.dictConfig(log_config)
                self.logger = logging.getLogger(str(self.__class__))
            if not os.path.isfile(os.path.join(self.logs_path, "project_log.log")):
                self.logger.info(f"Logging initiated in project {self.project_path}...")
        except:
            self.logger = None

    def create_third_part_append_logger(self, path: str) -> None:
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        handler = logging.FileHandler(filename=path, encoding="utf-8")  # or whatever
        handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
        logger.addHandler(handler)

    def add_video_to_video_info(
        self,
        video_name: str,
        fps: float,
        width: int,
        height: int,
        pixels_per_mm: int,
        distance_mm: float = 0.0,
    ) -> None:
        self.video_info_df.loc[len(self.video_info_df)] = [
            video_name,
            fps,
            width,
            height,
            distance_mm,
            pixels_per_mm,
        ]
        self.video_info_df.drop_duplicates(subset=["Video"], inplace=True, keep="last")
        self.video_info_df = self.video_info_df.set_index("Video")
        try:
            self.video_info_df.to_csv(
                os.path.join(self.project_path, Paths.VIDEO_INFO.value)
            )
        except:
            raise PermissionError(
                msg="SimBA tried to write to project_folder/logs/video_info.csv, but was not allowed. If this file is open in another program, try closing it.",
                source=self.__class__.__name__,
            )
        stdout_success(
            msg="Video info has been UPDATED at project_folder/logs/video_info.csv",
            source=self.__class__.__name__,
        )


# config = ConfigReader(config_path='/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/two_c57/project_folder/project_config.ini', read_video_info=False)

# config.read_roi_data()


# config = ConfigReader(config_path='/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/project_config.ini', read_video_info=False)
# config.read_roi_data()
# config.remove_roi_features('/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/csv/features_extracted')

# remove_roi_features(
#     config_path='/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/project_config.ini',
#     data_dir='/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/csv/features_extracted')

# test = ConfigReader(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



import argparse
import os
import sys
from typing import Union

import numpy as np
import pandas as pd

from simba.data_processors.find_animal_blob_location import (
    get_blob_vertices_from_video, get_left_right_points,
    get_nose_tail_from_vertices, stabilize_body_parts)
from simba.mixins.geometry_mixin import GeometryMixin
from simba.utils.checks import (check_if_dir_exists, check_instance, check_int,
                                check_nvidea_gpu_available, check_valid_dict)
from simba.utils.data import resample_geometry_vertices, savgol_smoother
from simba.utils.errors import SimBAGPUError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_core_cnt, get_video_meta_data,
                                    read_pickle, remove_files, write_df)
from simba.video_processors.video_processing import (video_bg_subtraction,
                                                     video_bg_subtraction_mp)

CENTER_X = 'center_x'
CENTER_Y = 'center_y'
NOSE_X = 'nose_x'
NOSE_Y = 'nose_y'
TAIL_X = 'tail_x'
TAIL_Y = 'tail_y'
LEFT_X = 'left_x'
LEFT_Y = 'left_y'
RIGHT_X = 'right_x'
RIGHT_Y = 'right_y'
VERTICES = 'vertices'

IN_DIR = 'input_dir'
OUT_DIR = 'output_dir'
VIDEO_DATA = 'video_data'
VERTICE_CNT = 'vertice_cnt'
REFERENCE = 'reference'
CORE_CNT = 'core_cnt'
CLOSING_KERNEL = 'close_kernel'
CLOSING_ITS = 'close_iterations'
OPENING_ITS = 'open_iterations'
OPENING_KERNEL = 'open_kernel'
SAVE_BG_VIDEOS = 'save_bg_videos'
GPU = 'gpu'

REQUIRED_KEYS = (IN_DIR, OUT_DIR, GPU, CORE_CNT, VIDEO_DATA, VERTICE_CNT, SAVE_BG_VIDEOS, CLOSING_ITS, OPENING_ITS)

METHOD = 'method'
THRESHOLD = 'threshold'


class BlobTrackingExecutor():

    """
    Perform animal blob tracking from video data using background subtraction, blob location detection,
    and geometry-based processing. The class handles the processing of video frames in parallel, tracks blob locations
    across frames, and saves the results to CSV files.

    :param Union[dict, str, os.PathLike] data: Path to a configuration file or a dictionary containing the necessary parameters for blob tracking. The configuration should include keys for the video data, input/output directories, and other processing settings.
    :param int vertice_cnt: The number of vertices to use when sampling the blob geometry. Default: 10.
    :param int batch_size: The batch size for blob location detection. Default: 4k.

    :example:
    >>> tracker = BlobTrackingExecutor(data=r"C:\troubleshooting\mitra\test\.temp\blob_definitions.h5", vertice_cnt=10)
    >>> tracker.run()
    """


    def __init__(self,
                 data: Union[dict, str, os.PathLike],
                 batch_size: int = 5000,
                 rostrocaudal : bool = True,
                 mediolateral: bool = True,
                 center: bool = True):

        self.timer = SimbaTimer(start=True)
        check_instance(source=f'{self.__class__.__name__} data', instance=data, accepted_types=(str, os.PathLike, dict), raise_error=True)
        if isinstance(data, (str, os.PathLike)):
            data = read_pickle(data_path=data)
        check_valid_dict(x=data, valid_key_dtypes=(str,), required_keys=REQUIRED_KEYS)
        check_if_dir_exists(in_dir=data[IN_DIR], source=self.__class__.__name__, raise_error=True)
        check_if_dir_exists(in_dir=data[OUT_DIR], source=self.__class__.__name__, raise_error=True)
        check_int(name=f'{self.__class__.__name__} core_cnt', value=data[CORE_CNT], min_value=1, max_value=find_core_cnt()[0], raise_error=True)
        check_int(name=f'{self.__class__.__name__} batch_size', value=batch_size, min_value=1, raise_error=True)
        self.data, self.gpu, self.core_cnt, self.batch_size, self.vertice_cnt, self.closing_iterations = data, data[GPU], data[CORE_CNT], batch_size, data[VERTICE_CNT], data[CLOSING_ITS]
        self.save_bg_videos, self.save_dir, self.opening_iterations = data[SAVE_BG_VIDEOS], data[OUT_DIR], data[OPENING_ITS]
        self.rostrocaudal, self.mediolateral, self.center = rostrocaudal, mediolateral, center
        if self.gpu and not check_nvidea_gpu_available():
            raise SimBAGPUError(msg='GPU is set to True, but SImBA could not find a GPU on the machine', source=self.__class__.__name__)
        self.vertice_col_names = []
        self.video_cnt = len(list(self.data[VIDEO_DATA].keys()))
        for i in range(self.vertice_cnt):
            self.vertice_col_names.append(f"vertice_{i}_x"); self.vertice_col_names.append(f"vertice_{i}_y")

    def run(self):
        #self._remove_bgs()
        self._find_blobs()
        self.timer.stop_timer()
        stdout_success(msg=f'Animal tracking complete. Results save din directory {self.data[OUT_DIR]}', elapsed_time=self.timer.elapsed_time_str)

    def _interpolate_vertices(self, arr: np.ndarray) -> np.ndarray:
        df = pd.DataFrame(arr.reshape(arr.shape[0], -1))
        df = df.fillna(0).clip(lower=0).astype(np.int32)
        for c in df.columns:
            bp_df = df[[c]].astype(np.float32)
            missing_idx = bp_df[bp_df[c] <= 0.0].index
            bp_df.loc[missing_idx, c] = np.nan
            if bp_df[c].notna().sum() >= 2:
                bp_df[c] = bp_df[c].interpolate(method='nearest', axis=0).ffill().bfill()
            else:
                bp_df[c] = bp_df[c].fillna(0)
            df.update(bp_df)
        return df.values.reshape(arr.shape).astype(np.int32)

    def _remove_bgs(self):
        bg_timer = SimbaTimer(start=True)
        self.bg_video_paths = []
        for video_cnt, (video_name, video_data) in enumerate(self.data[VIDEO_DATA].items()):
            video_meta = get_video_meta_data(video_path=video_data['video_path'])
            bg_video_path = os.path.join(self.save_dir, f'{video_name}.mp4')
            self.bg_video_paths.append(bg_video_path)
            close_kernel_size = None if video_data[CLOSING_KERNEL] is None else tuple(video_data[CLOSING_KERNEL])
            opening_kernel_size = None if video_data[OPENING_KERNEL] is None else tuple(video_data[OPENING_KERNEL])
            print(f'Starting background subtraction on video {video_name} (video {video_cnt+1}/{self.video_cnt}, frame count: {video_meta["frame_count"]}, cores: {self.core_cnt}, gpu: {self.gpu})')
            if (self.core_cnt == 1):
                video_bg_subtraction(video_path=video_data['video_path'],
                                     bg_video_path=video_data[REFERENCE],
                                     threshold=video_data[THRESHOLD],
                                     save_path=bg_video_path,
                                     verbose=True,
                                     fg_color=(255, 255, 255),
                                     closing_kernel_size=close_kernel_size,
                                     closing_iterations=self.closing_iterations,
                                     opening_kernel_size=opening_kernel_size,
                                     opening_iterations=self.opening_iterations)
            else:
                video_bg_subtraction_mp(video_path=video_data['video_path'],
                                        bg_video_path=video_data[REFERENCE],
                                        threshold=video_data[THRESHOLD],
                                        save_path=bg_video_path,
                                        verbose=True,
                                        gpu=self.gpu,
                                        core_cnt=self.core_cnt,
                                        fg_color=(255, 255, 255),
                                        closing_kernel_size=close_kernel_size,
                                        closing_iterations=self.closing_iterations,
                                        opening_kernel_size=opening_kernel_size,
                                        opening_iterations=self.opening_iterations)

        bg_timer.stop_timer()
        print(f'Background subtraction COMPLETE: videos saved at {self.save_dir}, (elapsed time: {bg_timer.elapsed_time_str}s)')



    def _find_blobs(self):
        blob_timer = SimbaTimer(start=True)
        for video_cnt, (video_name, video_data) in enumerate(self.data[VIDEO_DATA].items()):
            video_timer, geometries = SimbaTimer(start=True), None
            video_meta = get_video_meta_data(video_path=video_data['video_path'])
            temp_video_path = os.path.join(self.save_dir, f'{video_name}.mp4')
            save_path = os.path.join(self.data[OUT_DIR], f'{video_meta["video_name"]}.csv')
            inclusion_zone = None if 'inclusion_zones' not in video_data.keys() else video_data['inclusion_zones']
            window_size = None if 'window_size' not in video_data.keys() else video_data['window_size']
            vertices = get_blob_vertices_from_video(video_path=temp_video_path,
                                                    gpu=self.gpu,
                                                    verbose=True,
                                                    core_cnt=self.core_cnt,
                                                    batch_size=None,
                                                    inclusion_zone=inclusion_zone,
                                                    window_size=window_size,
                                                    convex_hull=False,
                                                    vertice_cnt=self.vertice_cnt)

            vertices = self._interpolate_vertices(arr=vertices)
            results = pd.DataFrame()
            if video_data['buffer_size'] is not None:
                geometries = GeometryMixin.bodyparts_to_polygon(data=vertices, parallel_offset=video_data['buffer_size'], convex_hull=False, simplify_tolerance=0.1)
                vertices = [np.array(x.exterior.coords).astype(np.int32) for x in geometries]
                vertices = resample_geometry_vertices(vertices=vertices, vertice_cnt=self.vertice_cnt)
            if video_data['smoothing_time'] is not None:
                vertices = savgol_smoother(data=vertices.reshape(vertices.shape[0], -1), fps=video_meta['fps'], time_window=video_data['smoothing_time'], source=self.__class__.__name__).astype(np.int32)
                vertices = vertices.reshape(vertices.shape[0], self.vertice_cnt, 2)
            if self.center or self.mediolateral or self.rostrocaudal:
                geometries = geometries if geometries != None else GeometryMixin.bodyparts_to_polygon(data=vertices, convex_hull=False, simplify_tolerance=0.1)
                centers = GeometryMixin.get_center(shape=geometries).astype(np.int32)
                results[CENTER_X], results[CENTER_Y] = centers[:, 0], centers[:, 1]
            if self.rostrocaudal or self.mediolateral:
                nose, tail = get_nose_tail_from_vertices(vertices=vertices, fps=video_meta['fps'], smooth_factor=1.0)
                #nose, tail = stabilize_body_parts(bp_1=nose, bp_2=tail, center_positions=centers)
                results[NOSE_X], results[NOSE_Y] = nose[:, 0], nose[:, 1]
                results[TAIL_X], results[TAIL_Y] = tail[:, 0], tail[:, 1]
            if self.mediolateral:
                left, right = get_left_right_points(hull_vertices=vertices, anterior=nose, center=centers, posterior=tail)
                results[LEFT_X], results[LEFT_Y] = left[:, 0], left[:, 1]
                results[RIGHT_X], results[RIGHT_Y] = right[:, 0], right[:, 1]

            vertices = pd.DataFrame(vertices.reshape(vertices.shape[0], (self.vertice_cnt*2)).astype(np.int32), columns=self.vertice_col_names)
            results = pd.concat([results, vertices], axis=1).reset_index(drop=True).fillna(0)
            write_df(df=results, file_type='csv', save_path=save_path)
            video_timer.stop_timer()
            print(f'Animal blob tracking data for video {video_meta["video_name"]} saved at {save_path}, (elapsed time: {video_timer.elapsed_time_str}s).')
        blob_timer.stop_timer()
        if not self.save_bg_videos:
            remove_files(file_paths=self.bg_video_paths, raise_error=False)
        print(f'Blob tracking COMPLETE: data saved at {self.save_dir}, (elapsed time: {blob_timer.elapsed_time_str}s')

if __name__ == "__main__" and not hasattr(sys, 'ps1'):
    parser = argparse.ArgumentParser(description="Execute Blob tracking in SimBA.")
    parser.add_argument('--data', type=str, required=True, help='Path to the pickle holding the parameters for performing blob tracking')
    args = parser.parse_args()
    tracker = BlobTrackingExecutor(data=args.data)
    tracker.run()


# DATA_PATH = r"C:\troubleshooting\blob_track_tester\results\blob_definitions.pickle"
# tracker = BlobTrackingExecutor(data=DATA_PATH)
# tracker.run()



# #DATA_PATH = r"/mnt/d/open_field_3/sample/.temp/blob_definitions.h5"
# #DATA_PATH = r"D:\open_field_3\sample\blob_data\blob_definitions.json"
# # DATA_PATH = r"D:\EPM\sampled\.temp\blob_definitions.h5"
# #DATA_PATH = r"D:\EPM\sample_2\.temp\blob_definitions.h5"
#
# #
# DATA_PATH = r"D:\open_field_2\sample\clipped_10min\data\blob_definitions.json"
#
#
# #DATA_PATH = r"D:\open_field_4\data\blob_definitions.json"
# DATA_PATH = r"D:\open_field\data\blob_definitions.json"
#
# DATA_PATH = r"D:\EPM_3\out\blob_definitions.json"
# # # # # # # # # #
# # # # # # # # # #
# tracker = BlobTrackingExecutor(data=DATA_PATH)
# tracker.run()
# # # #
# # #
# #

# data = {'input_dir': r'C:\\troubleshooting\\mitra\\test',
#         'output_dir': r'C:\\troubleshooting\\mitra\\test\\blob_data',
#         'gpu': False,
#         'core_cnt': 32,
#         'video_data': {r'C:\\troubleshooting\\mitra\\test\\501_MA142_Gi_Saline_0515.mp4': {'threshold': 50, 'method': 'absolute', 'window_size': 'None', 'window_weight': 'None', 'reference': r'C:/troubleshooting/mitra/test\background_dir\501_MA142_Gi_Saline_0515.mp4', 'visualize': False, 'exclusion_zones': None,
#                                                                                            'inclusion_zones': {'inclusion_zone': {'Video': '501_MA142_Gi_Saline_0515', 'Shape_type': 'rectangle', 'Name': 'inclusion_zone', 'Color name': 'Red', 'Color BGR': (0, 0, 255), 'Thickness': 5, 'Center_X': 365, 'Center_Y': 293, 'topLeftX': 204, 'topLeftY': 131, 'Bottom_right_X': 526, 'Bottom_right_Y': 455, 'width': 322, 'height': 324, 'width_cm': 32.2, 'height_cm': 32.4, 'area_cm': 1043.28, 'Tags': {'Center tag': (365, 293), 'Top left tag': (204, 131), 'Bottom right tag': (526, 455), 'Top right tag': (526, 131), 'Bottom left tag': (204, 455), 'Top tag': (365, 131), 'Right tag': (526, 293), 'Left tag': (204, 293), 'Bottom tag': (365, 455)}, 'Ear_tag_size': 15}}}}}
#



# data = {'input_dir': r'C:\\troubleshooting\\mitra\\test',
#         'output_dir': r'C:\\troubleshooting\\mitra\\test\\blob_data',
#         'gpu': False,
#         'core_cnt': 32,
#         'video_data': {r'C:\\troubleshooting\\mitra\\test\\501_MA142_Gi_Saline_0515': {'threshold': 30, 'method': 'absolute', 'window_size': 'None', 'window_weight': 'None', 'reference': 'C:/troubleshooting/mitra/test\\501_MA142_Gi_Saline_0515.mp4', 'visualize': False, 'exclusion_zones': None, 'inclusion_zones': None},
#                        r'C:\\troubleshooting\\mitra\\test\\502_MA141_Gi_CNO_0514': {'threshold': 30, 'method': 'absolute', 'window_size': 'None', 'window_weight': 'None', 'reference': 'C:/troubleshooting/mitra/test\\502_MA141_Gi_CNO_0514.mp4', 'visualize': False, 'exclusion_zones': None, 'inclusion_zones': None},
#                        r'C:\\troubleshooting\\mitra\\test\\503_MA109_Gi_CNO_0521': {'threshold': 30, 'method': 'absolute', 'window_size': 'None', 'window_weight': 'None', 'reference': 'C:/troubleshooting/mitra/test\\503_MA109_Gi_CNO_0521.mp4', 'visualize': False, 'exclusion_zones': None, 'inclusion_zones': None}}}
#



# data = read_pickle(data_path=r"C:\troubleshooting\mitra\blob_data.pickle")
# # data['gpu'] = True
# # data['core_cnt'] = 16
# tracker = BlobTrackingExecutor(data=data)
# tracker.run()


__author__ = "Simon Nilsson"

import itertools
import os
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.roi_tools.roi_aggregate_statistics_analyzer import \
    ROIAggregateStatisticsAnalyzer
from simba.roi_tools.roi_utils import get_roi_dict_from_dfs
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict,
                                check_if_valid_rgb_tuple, check_int,
                                check_valid_boolean, check_valid_lst,
                                check_video_and_data_frm_count_align)
from simba.utils.data import (create_color_palettes, detect_bouts,
                              slice_roi_dict_for_video)
from simba.utils.enums import Formats, Keys, Paths, TagNames, TextOptions
from simba.utils.errors import (BodypartColumnNotFoundError, DuplicationError,
                                NoFilesFoundError, NoROIDataError,
                                ROICoordinatesNotFoundError)
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (get_video_meta_data, read_df,
                                    read_frm_of_video)
from simba.utils.warnings import FrameRangeWarning

SHOW_BODY_PARTS = 'show_body_part'
SHOW_ANIMAL_NAMES = 'show_animal_name'
STYLE_KEYS = [SHOW_BODY_PARTS, SHOW_ANIMAL_NAMES]
OUTSIDE_ROI = 'OUTSIDE REGIONS OF INTEREST'


class ROIPlotter(ConfigReader):
    """
    Visualize the ROI data (number of entries/exits, time-spent in ROIs etc).

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    .. seelalso::
       Use :func:`simba.plotting.ROI_plotter_mp.ROIPlotMultiprocess` for improved run-time.

    .. image:: _static/img/ROIPlot_1.png
       :width: 800
       :align: center

    .. image:: _static/img/ROIPlot_1.gif
       :width: 800
       :align: center

    .. video:: _static/img/outside_roi_example.mp4
       :width: 800
       :autoplay:
       :loop:

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] video_path: Name of video to create ROI visualizations for
    :param Dict[str, bool] style_attr: User-defined visualization settings.
    :param List[str] body_parts: List of the body-parts to use as proxy for animal locations.
    :param Optional[float] threshold: Float between 0 and 1. Body-part locations detected below this confidence threshold are filtered. Default: 0.0.
    :param Optional[bool]: If True, SimBA will treat all areas NOT covered by a ROI drawing as a single additional ROI visualize the stats for this, single, ROI.
    :param Optional[Union[str, os.PathLike]] data_path: Optional path to the pose-estimation data. If None, then locates file in ``outlier_corrected_movement_location`` directory.
    :param Optional[Union[str, os.PathLike]] save_path: Optional path to where to save video If None, then saves it in ''frames/output/roi_analys`` directory of SimBA project.
    :param Optional[List[Tuple[int, int, int]] bp_colors: Optional list of tuples of same length as body_parts representing the colors of the body-parts. Defaults to None and colors are automatically chosen.
    :param Optional[List[int]] bp_sizes: Optional list of integers representing the sizes of the pose estimated body-part location. Defaults to None and size is automnatically inferred.


    :example:
    >>> test = ROIPlotter(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/project_config.ini',
    >>>                video_path="/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/videos/SI_DAY3_308_CD1_PRESENT.mp4",
    >>>                body_parts=['Nose'],
    >>>                style_attr={'show_body_part': True, 'show_animal_name': True})
    >>> test.run()


    :example II:
    >>> test = ROIPlotter(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
    >>>               video_path=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_Saline_0513.mp4",
    >>>               body_parts=['Nose'],
    >>>               style_attr={'show_body_part': True, 'show_animal_name': False})
    >>> test.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 video_path: Union[str, os.PathLike],
                 style_attr: Dict[str, bool],
                 body_parts: List[str],
                 outside_roi: bool = False,
                 threshold: float = 0.0,
                 verbose: Optional[bool] = True,
                 data_path: Optional[Union[str, os.PathLike]] = None,
                 save_path: Optional[Union[str, os.PathLike]] = None,
                 bp_colors: Optional[List[Tuple[int, int, int]]] = None,
                 bp_sizes: Optional[List[Union[int]]] = None):

        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=0.0, max_value=1.0)
        check_if_keys_exist_in_dict(data=style_attr, key=STYLE_KEYS, name=f'{self.__class__.__name__} style_attr')
        check_valid_boolean(value=outside_roi, source=f'{self.__class__.__name__} outside_roi', raise_error=True)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        self.video_meta = get_video_meta_data(video_path=video_path)
        self.video_path = video_path
        self.video_name = self.video_meta['video_name']
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        self.read_roi_data()
        self.sliced_roi_dict, self.shape_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=self.video_name)
        if len(self.shape_names) == 0:
            raise NoROIDataError(msg=f"Cannot plot ROI data for video {self.video_name}. No ROIs defined for this video.")
        if data_path is None:
            data_path = os.path.join(self.outlier_corrected_dir, f'{self.video_name}.{self.file_type}')
        else:
            if not os.path.isfile(data_path):
                raise NoFilesFoundError(msg=f"SIMBA ERROR: Could not find the file at path {data_path}. Make sure the data file exist to create ROI visualizations",  source=self.__class__.__name__)
            check_file_exist_and_readable(file_path=data_path)
        if save_path is None:
            save_path = os.path.join(self.project_path, Paths.ROI_ANALYSIS.value, f'{self.video_name}.mp4')
            if not os.path.exists(os.path.dirname(save_path)): os.makedirs(os.path.dirname(save_path))
        else:
            check_if_dir_exists(os.path.dirname(save_path))
        self.save_path, self.data_path = save_path, data_path
        check_valid_lst(data=body_parts, source=f'{self.__class__.__name__} body-parts', valid_dtypes=(str,), min_len=1)
        if outside_roi: self.shape_names.append(OUTSIDE_ROI)
        if len(set(body_parts)) != len(body_parts):
            raise DuplicationError(msg=f'All body-part entries have to be unique. Got {body_parts}', source=self.__class__.__name__)
        for bp in body_parts:
            if bp not in self.body_parts_lst: raise BodypartColumnNotFoundError(msg=f'The body-part {bp} is not a valid body-part in the SimBA project. Options: {self.body_parts_lst}', source=self.__class__.__name__)
        self.roi_analyzer = ROIAggregateStatisticsAnalyzer(config_path=self.config_path, data_path=self.data_path,  detailed_bout_data=True, threshold=threshold, body_parts=body_parts, outside_rois=outside_roi, verbose=verbose)
        self.roi_analyzer.run()
        if bp_colors is not None:
            check_valid_lst(data=bp_colors, source=f'{self.__class__.__name__} bp_colors', valid_dtypes=(tuple,), exact_len=len(body_parts), raise_error=True)
            _ = [check_if_valid_rgb_tuple(x) for x in bp_colors]
            self.color_lst = bp_colors
        else:
            self.color_lst = create_color_palettes(self.roi_analyzer.animal_cnt, len(body_parts))[0]
        self.bp_sizes = bp_sizes
        try:
            self.detailed_roi_data = pd.concat(self.roi_analyzer.detailed_dfs, axis=0).reset_index(drop=True)
        except ValueError:
            self.detailed_roi_data = None

        self.detailed_roi_data = pd.concat(self.roi_analyzer.detailed_dfs, axis=0).reset_index(drop=True)
        self.bp_dict = self.roi_analyzer.bp_dict
        self.animal_names = [self.find_animal_name_from_body_part_name(bp_name=x, bp_dict=self.animal_bp_dict) for x in body_parts]
        self.data_df = read_df(file_path=self.data_path, file_type=self.file_type, usecols=self.roi_analyzer.roi_headers).fillna(0.0).reset_index(drop=True)
        self.shape_columns = []
        for x in itertools.product(self.animal_names, self.shape_names):
            self.data_df[f"{x[0]}_{x[1]}"] = 0; self.shape_columns.append(f"{x[0]}_{x[1]}")
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        check_video_and_data_frm_count_align(video=self.video_path, data=self.data_df, name=self.video_name, raise_error=False)
        self.cap = cv2.VideoCapture(self.video_path)
        self.threshold, self.body_parts, self.style_attr, self.outside_roi, self.verbose = threshold, body_parts, style_attr, outside_roi, verbose
        self.roi_dict_ = get_roi_dict_from_dfs(rectangle_df=self.sliced_roi_dict[Keys.ROI_RECTANGLES.value], circle_df=self.sliced_roi_dict[Keys.ROI_CIRCLES.value], polygon_df=self.sliced_roi_dict[Keys.ROI_POLYGONS.value])

    def __get_circle_sizes(self):
        optimal_circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(int(self.video_meta["height"]), int(self.video_meta["height"])), circle_frame_ratio=70)
        if self.bp_sizes is None:
            self.circle_sizes = [optimal_circle_size] * len(self.animal_names)
        else:
            self.circle_sizes = []
            for circle_size in self.bp_sizes:
                if not check_int(name='circle_size', value=circle_size, min_value=1, raise_error=False)[0]:
                    self.circle_sizes.append(optimal_circle_size)
                else:
                    self.circle_sizes.append(int(circle_size))

    def __get_roi_columns(self):
        if self.detailed_roi_data is not None:
            roi_entries_dict = self.detailed_roi_data[["ANIMAL", "Event", "Start_frame", "End_frame"]].to_dict(orient="records")
            for entry_dict in roi_entries_dict:
                entry, exit = int(entry_dict["Start_frame"]), int(entry_dict["End_frame"])
                entry_dict["frame_range"] = list(range(entry, exit + 1))
                col_name =  f'{entry_dict["ANIMAL"]}_{entry_dict["Event"]}'
                self.data_df[col_name][self.data_df.index.isin(entry_dict["frame_range"])] = 1

    def __get_bordered_img_size(self) -> Tuple[int, int]:
        img = read_frm_of_video(video_path=self.video_path, frame_index=0)
        self.base_img = cv2.copyMakeBorder(img, 0, 0, 0, int(self.video_meta["width"]), borderType=cv2.BORDER_CONSTANT, value=[0, 0, 0])
        return (self.base_img.shape[0], self.base_img.shape[1])

    def __get_text_locs(self) -> dict:
         loc_dict = {}
         txt_strs = []
         for animal_cnt, animal_name in enumerate(self.animal_names):
             for shape in self.shape_names:
                 txt_strs.append(f'{animal_name} {shape} entries')
         longest_text_str = max(txt_strs, key=len)
         self.font_size, x_spacer, y_spacer = PlottingMixin().get_optimal_font_scales(text=longest_text_str, accepted_px_width=int(self.video_meta["width"] / 1.5), accepted_px_height=int(self.video_meta["height"] / 10), text_thickness=TextOptions.TEXT_THICKNESS.value)
         row_counter = TextOptions.FIRST_LINE_SPACING.value
         for animal_cnt, animal_name in enumerate(self.animal_names):
             loc_dict[animal_name] = {}
             for shape in self.shape_names:
                 loc_dict[animal_name][shape] = {}
                 loc_dict[animal_name][shape]["timer_text"] = f"{shape} {animal_name} timer:"
                 loc_dict[animal_name][shape]["entries_text"] = f"{shape} {animal_name} entries:"
                 loc_dict[animal_name][shape]["timer_text_loc"] = ((self.video_meta["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta["height"] - (self.video_meta["height"] + TextOptions.BORDER_BUFFER_Y.value) + y_spacer * row_counter))
                 loc_dict[animal_name][shape]["timer_data_loc"] = (int(self.video_meta["width"] + x_spacer + TextOptions.BORDER_BUFFER_X.value), (self.video_meta["height"] - (self.video_meta["height"] + TextOptions.BORDER_BUFFER_Y.value) + y_spacer * row_counter))
                 row_counter += 1
                 loc_dict[animal_name][shape]["entries_text_loc"] = ((self.video_meta["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta["height"] - (self.video_meta["height"] + TextOptions.BORDER_BUFFER_Y.value) + y_spacer * row_counter))
                 loc_dict[animal_name][shape]["entries_data_loc"] = (int(self.video_meta["width"] + x_spacer + TextOptions.BORDER_BUFFER_X.value), (self.video_meta["height"]- (self.video_meta["height"] + TextOptions.BORDER_BUFFER_Y.value) + y_spacer * row_counter))
                 row_counter += 1
         return loc_dict

    def __get_counters(self) -> dict:
        cnt_dict = {}
        for animal_cnt, animal_name in enumerate(self.animal_names):
            cnt_dict[animal_name] = {}
            for shape in self.shape_names:
                cnt_dict[animal_name][shape] = {}
                cnt_dict[animal_name][shape]["timer"] = 0
                cnt_dict[animal_name][shape]["entries"] = 0
                cnt_dict[animal_name][shape]["entry_status"] = False
        return cnt_dict

    def __insert_texts(self, roi_dict, img):
        for animal_name in self.animal_names:
            for shape_name, shape_data in roi_dict.items():
                img = cv2.putText(img, self.loc_dict[animal_name][shape_name]["timer_text"], self.loc_dict[animal_name][shape_name]["timer_text_loc"], TextOptions.FONT.value, self.font_size, shape_data['Color BGR'], TextOptions.TEXT_THICKNESS.value)
                img = cv2.putText(img, self.loc_dict[animal_name][shape_name]["entries_text"], self.loc_dict[animal_name][shape_name]["entries_text_loc"], TextOptions.FONT.value, self.font_size, shape_data['Color BGR'], TextOptions.TEXT_THICKNESS.value)
            if self.outside_roi:
                img = cv2.putText(img, self.loc_dict[animal_name][OUTSIDE_ROI]["timer_text"], self.loc_dict[animal_name][OUTSIDE_ROI]["timer_text_loc"], TextOptions.FONT.value, self.font_size, TextOptions.WHITE.value, TextOptions.TEXT_THICKNESS.value)
                img = cv2.putText(img, self.loc_dict[animal_name][OUTSIDE_ROI]["entries_text"], self.loc_dict[animal_name][OUTSIDE_ROI]["entries_text_loc"], TextOptions.FONT.value, self.font_size, TextOptions.WHITE.value, TextOptions.TEXT_THICKNESS.value)
        return img

    def __get_cumulative_data(self):
        for animal_name in self.animal_names:
            for shape in self.shape_names:
                self.data_df[f"{animal_name}_{shape}_cum_sum_time"] = (self.data_df[f"{animal_name}_{shape}"].cumsum() / self.video_meta['fps'])
                roi_bouts = list(detect_bouts(data_df=self.data_df, target_lst=[f"{animal_name}_{shape}"], fps=self.video_meta['fps'])["Start_frame"])
                self.data_df[f"{animal_name}_{shape}_entry"] = 0
                self.data_df.loc[roi_bouts, f"{animal_name}_{shape}_entry"] = 1
                self.data_df[f"{animal_name}_{shape}_cum_sum_entries"] = (self.data_df[f"{animal_name}_{shape}_entry"].cumsum())

    def run(self):
        video_timer = SimbaTimer(start=True)
        self.__get_circle_sizes()
        self.__get_roi_columns()
        self.border_img_h, self.border_img_w = self.__get_bordered_img_size()
        writer = cv2.VideoWriter(self.save_path, self.fourcc, self.video_meta["fps"], (self.border_img_w, self.border_img_h))
        self.loc_dict = self.__get_text_locs()
        self.cnt_dict = self.__get_counters()
        self.__get_cumulative_data()
        frame_cnt = 0
        while self.cap.isOpened():
            ret, img = self.cap.read()
            if ret:
                img = cv2.copyMakeBorder(img, 0, 0, 0, int(self.video_meta["width"]),  borderType=cv2.BORDER_CONSTANT, value=[0, 0, 0])
                img = self.__insert_texts(self.roi_dict_, img)
                img = PlottingMixin.roi_dict_onto_img(img=img, roi_dict=self.sliced_roi_dict)
                for animal_cnt, animal_name in enumerate(self.animal_names):
                    x, y, p = (self.data_df.loc[frame_cnt, self.bp_dict[animal_name]].fillna(0.0).values.astype(np.int32))
                    if (self.threshold <= p) and self.style_attr[SHOW_BODY_PARTS]:
                        img = cv2.circle(img, (x, y), self.circle_sizes[animal_cnt], self.color_lst[animal_cnt], -1)
                    if (self.threshold <= p) and self.style_attr[SHOW_ANIMAL_NAMES]:
                        img = cv2.putText(img, animal_name, (x, y), self.font, self.font_size, self.color_lst[animal_cnt], TextOptions.TEXT_THICKNESS.value)
                for animal_cnt, animal_name in enumerate(self.animal_names):
                    for shape_name, shape_data in self.roi_dict_.items():
                        time = str(round(self.data_df.loc[frame_cnt, f"{animal_name}_{shape_name}_cum_sum_time"], 2))
                        entries = str(int(self.data_df.loc[frame_cnt, f"{animal_name}_{shape_name}_cum_sum_entries"]))
                        img = cv2.putText(img, time, self.loc_dict[animal_name][shape_name]["timer_data_loc"], self.font, self.font_size, shape_data["Color BGR"], TextOptions.TEXT_THICKNESS.value)
                        img = cv2.putText(img, entries, self.loc_dict[animal_name][shape_name]["entries_data_loc"], self.font, self.font_size, shape_data["Color BGR"], TextOptions.TEXT_THICKNESS.value)
                    if self.outside_roi:
                        time = str(round(self.data_df.loc[frame_cnt, f"{animal_name}_{OUTSIDE_ROI}_cum_sum_time"], 2))
                        entries = str(int(self.data_df.loc[frame_cnt, f"{animal_name}_{OUTSIDE_ROI}_cum_sum_entries"]))
                        img = cv2.putText(img, time, self.loc_dict[animal_name][OUTSIDE_ROI]["timer_data_loc"], self.font, self.font_size, TextOptions.WHITE.value, TextOptions.TEXT_THICKNESS.value)
                        img = cv2.putText(img, entries, self.loc_dict[animal_name][OUTSIDE_ROI]["entries_data_loc"], self.font, self.font_size, TextOptions.WHITE.value, TextOptions.TEXT_THICKNESS.value)
                writer.write(img)
                if self.verbose: print(f"Frame: {frame_cnt+1} / {self.video_meta['frame_count']}, Video: {self.video_name}.")
                frame_cnt += 1
            else:
                FrameRangeWarning(msg=f'Could not read frame {frame_cnt} in video {self.video_name}', source=self.__class__.__name__)
                break
        writer.release()
        video_timer.stop_timer()
        if self.verbose: stdout_success(msg=f"Video {self.video_name} created. Video saved at {self.save_path}", elapsed_time=video_timer.elapsed_time_str, source=self.__class__.__name__)



# if __name__ == "__main__":
#     test = ROIPlotter(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                video_path=r"C:\troubleshooting\mitra\project_folder\videos\502_MA141_Gi_Saline_0517.mp4",
#                                body_parts=['Nose'],
#                                style_attr={'show_body_part': True, 'show_animal_name': False})
#     test.run()


# test = ROIPlotter(config_path=r"C:\troubleshooting\roi_duplicates\project_folder\project_config.ini",
#                   video_path=r"C:\troubleshooting\roi_duplicates\project_folder\videos\2021-12-21_15-03-57_CO_Trimmed.mp4",
#                   body_parts=['Snout'],
#                   style_attr={'show_body_part': True, 'show_animal_name': False})
# test.run()



# if __name__ == "__main__":
#     test = ROIPlotMultiprocess(config_path=r"C:\troubleshooting\roi_duplicates\project_folder\project_config.ini",
#                                video_path=r"C:\troubleshooting\roi_duplicates\project_folder\videos\2021-12-21_15-03-57_CO_Trimmed.mp4",
#                                body_parts=['Snout'],
#                                style_attr={'show_body_part': True, 'show_animal_name': False},
#                                bp_sizes=[20],
#                                bp_colors=[(155, 255, 243)])
#     test.run()
#


# test = ROIPlotter(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                   video_path=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_Saline_0513.mp4",
#                   body_parts=['Nose'],
#                   style_attr={'show_body_part': True, 'show_animal_name': False},
#                   outside_roi=True)
# test.run()


# test = ROIPlot(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/project_config.ini',
#                video_path="/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/videos/SI_DAY3_308_CD1_PRESENT.mp4",
#                body_parts=['Nose'],
#                style_attr={'show_body_part': True, 'show_animal_name': True})
# test.run()

# test = ROIPlot(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                video_path="termite_test.mp4",
#                style_attr={'Show_body_part': True, 'Show_animal_name': True})
# test.insert_data()
# test.visualize_ROI_data()


# test = ROIPlot(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini', video_path="termite_test.mp4")
# test.insert_data()
# test.visualize_ROI_data()

# test = ROIPlot(ini_path=r'/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini', video_path=r"Together_1.avi")
# test.insert_data()
# test.visualize_ROI_data()


# test = ROIPlot(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                video_path="Together_1.avi",
#                style_attr={'Show_body_part': True, 'Show_animal_name': False},
#                body_parts={f'Simon': 'Ear_left_1'})
# test.insert_data()
# test.run()

# test = ROIPlot(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                video_path="termite_test.mp4",
#                style_attr={'Show_body_part': True, 'Show_animal_name': True},
#                body_parts={f'Simon': 'Termite_1_Head_1'})
# test.insert_data()
# test.run()



import argparse
import functools
import multiprocessing
import os
from copy import copy
from typing import Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.data_processors.light_dark_box_analyzer import LightDarkBoxAnalyzer
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_float, check_if_dir_exists, check_str,
                                check_video_and_data_frm_count_align)
from simba.utils.enums import Defaults, Formats, Options, TextOptions
from simba.utils.errors import NoDataError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data)


def _light_dark_box_visualizer(pose_data: pd.DataFrame,
                               light_data: pd.DataFrame,
                               video_temp_dir: str,
                               video_path: str,
                               circle_size: Union[int, float],
                               txt_shift: Tuple[int, int],
                               font_size: Union[int, float],
                               body_part: str,
                               threshold: float):


    group, idx = int(pose_data[0]), pose_data[1].index.tolist()
    start_frm, end_frm = idx[0], idx[-1]
    video_meta_data = get_video_meta_data(video_path=video_path)
    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    video_save_path = os.path.join(video_temp_dir, f"{group}.mp4")
    video_writer = cv2.VideoWriter(video_save_path, fourcc, video_meta_data["fps"], (video_meta_data["width"], video_meta_data["height"]))
    cap = cv2.VideoCapture(video_path)
    light_data = light_data[light_data['EVENT'] == 'LIGHT BOX'].reset_index(drop=True)
    light_data['FRAME LIST'] = light_data.apply(lambda row: list(range(row['START FRAME'], row['END FRAME'] + 1)), axis=1)
    light_box_frm_lst = [f for s in light_data['FRAME LIST'] for f in s]
    for frm_cnt, frm_id in enumerate(range(start_frm, end_frm + 1)):
        line_counter, img_text_shift = 1, copy(txt_shift)
        cap.set(1, int(frm_id))
        ret, img = cap.read()
        x, y, p = pose_data[1].loc[frm_id].values
        if frm_id in light_box_frm_lst:
            img = cv2.circle(img, (int(x), int(y)), circle_size, (0, 0, 255), -1)
            img = PlottingMixin().put_text(img=img, text='ZONE: LIGHT BOX', pos=(TextOptions.BORDER_BUFFER_Y.value, int(line_counter * txt_shift[1])), font_size=font_size, text_color=(0, 255, 255), text_bg_alpha=0.8)
            line_counter += 1
            img_text_shift = (TextOptions.BORDER_BUFFER_Y.value, int(txt_shift[1] * line_counter))
            img = PlottingMixin().put_text(img=img, text=f'BODY-PART PROBABILITY: {round(p, 4)}', pos=img_text_shift, font_size=font_size, text_color=(0, 255, 255), text_bg_alpha=0.8)
        else:
            img = PlottingMixin().put_text(img=img, text='ZONE: DARK BOX', pos=(TextOptions.BORDER_BUFFER_Y.value, int(line_counter * txt_shift[1])), font_size=font_size, text_color=(255, 255, 255), text_bg_alpha=0.8)
            line_counter += 1
            img_text_shift = (TextOptions.BORDER_BUFFER_Y.value, int(txt_shift[1] * line_counter))
            img = PlottingMixin().put_text(img=img, text=f'BODY-PART PROBABILITY: {round(p, 4)}', pos=img_text_shift, font_size=font_size, text_color=(255, 255, 255), text_bg_alpha=0.8)
        line_counter += 1
        img_text_shift = (TextOptions.BORDER_BUFFER_Y.value, int(txt_shift[1] * line_counter))
        img = PlottingMixin().put_text(img=img, text=f'BODY-PART: {body_part}', pos=img_text_shift, font_size=font_size, text_color=(255, 255, 255), text_bg_alpha=0.8)
        line_counter += 1
        img_text_shift = (TextOptions.BORDER_BUFFER_Y.value, int(txt_shift[1] * line_counter))
        img = PlottingMixin().put_text(img=img, text=f'THRESHOLD: {threshold}', pos=img_text_shift, font_size=font_size, text_color=(255, 255, 255), text_bg_alpha=0.8)
        video_writer.write(img.astype(np.uint8))
        print(f"Multi-processing video frame {frm_id} on core {group}...")
        #break

    cap.release()
    video_writer.release()
    return group



class LightDarkBoxPlotter():

    """
    Generate annotated videos visualizing behavior episodes in a light/dark box setup.

    .. seealso::
       For light/dark box data analysis, see :func:`simba.data_processors.light_dark_box_analyzer.LightDarkBoxAnalyzer`.

    .. video:: _static/img/LightDarkBoxPlotter.webm
       :width: 800
       :autoplay:
       :loop:

    :param (str or os.PathLike) data_dir: Directory containing pose estimation CSV files.
    :param (str or os.PathLike) video_dir: Directory containing video files corresponding with the names of the CSVs in ``data_dir``.
    :param (str or os.PathLike) save_dir: Output directory to save the resulting annotated videos.
    :param (str) body_part: The name of the body part used to determine position and behavior.
    :param (int or float) fps : Frames per second of the videos (used for timing episodes).
    :param (float) threshold: Threshold value for light/dark classification (between 0.0 and 1.0).
    :param (float) minimum_episode_duration : Minimum duration (in seconds) for an episode to be considered valid.
    :param (int) core_cnt: Number of CPU cores to use for parallel processing. If -1, uses all available cores.

    :references:
       .. [1] For discussion about the development, see - `GitHub issue 446 <https://github.com/sgoldenlab/simba/issues/446#issuecomment-2930692735>`_.
    """

    def __init__(self,
                 video_dir: Union[str, os.PathLike],
                 data_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 body_part: str,
                 fps: Union[int, float],
                 threshold: float = 0.01,
                 minimum_episode_duration: float = 10e-16,
                 core_cnt: int = -1):

        check_if_dir_exists(in_dir=data_dir, source=f'{self.__class__.__name__} data_dir', raise_error=True)
        check_if_dir_exists(in_dir=video_dir, source=f'{self.__class__.__name__} video_dir', raise_error=True)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir', raise_error=True)
        print(data_dir)
        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'], raise_error=True, as_dict=True)
        self.video_paths = find_files_of_filetypes_in_directory(directory=video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True, as_dict=True)
        missing_videos = [x for x in self.data_paths.keys() if x not in self.video_paths.keys()]
        if len(missing_videos) > 0:
            raise NoDataError(msg=f'Video(s) {missing_videos} are missing from the video directory {video_dir}', source=self.__class__.__name__)
        self.data_dir = data_dir
        check_str(name=f'{self.__class__.__name__} body_part', value=body_part)
        check_float(name=f'{self.__class__.__name__} fps', value=fps, min_value=10e-6)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=0.0, max_value=1.0)
        check_float(name=f'{self.__class__.__name__} minimum_duration', value=minimum_episode_duration, min_value=0.0)
        self.body_part, self.fps, self.threshold, self.min_dur, self.core_cnt = body_part, fps, threshold, minimum_episode_duration, core_cnt
        self.save_dir = save_dir
        self.timer = SimbaTimer(start=True)

    def run(self):
        analyzer = LightDarkBoxAnalyzer(data_dir=self.data_dir, save_path=None, body_part=self.body_part, fps=self.fps, threshold=self.threshold, minimum_episode_duration=self.min_dur)
        for file_cnt, (file_name, file_path) in enumerate(self.data_paths.items()):
            video_timer = SimbaTimer(start=True)
            analyzer.data_paths = {file_name: file_path}
            analyzer.run()
            behavior_data = analyzer.results[analyzer.results['VIDEO'] == file_name].reset_index(drop=True)
            pose_data = analyzer.data_df[analyzer.bp_cols]
            pose_data[f'{self.body_part}_p'] = analyzer.bp_p
            check_video_and_data_frm_count_align(video=self.video_paths[file_name], data=pose_data, name=file_name, raise_error=True)
            pose_data = np.array_split(pose_data, self.core_cnt)
            pose_data = [(c, x) for c, x in enumerate(pose_data)]
            self.temp_dir = os.path.join(self.save_dir, file_name, "temp")
            self.video_save_path = os.path.join(self.save_dir, f'{file_name}.mp4')
            if not os.path.isdir(self.temp_dir): os.makedirs(self.temp_dir)
            video_meta = get_video_meta_data(video_path=self.video_paths[file_name])
            circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(video_meta['width'], video_meta['height']), circle_frame_ratio=50)
            font_size, x_shift, y_shift = PlottingMixin().get_optimal_font_scales(text='LONG ARSE STRING BODY-PART', accepted_px_width=int(video_meta['width']/2), accepted_px_height=int(video_meta['height']/2))
            with multiprocessing.Pool(self.core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value) as pool:
                constants = functools.partial(_light_dark_box_visualizer,
                                              light_data=behavior_data,
                                              video_path=self.video_paths[file_name],
                                              video_temp_dir=self.temp_dir,
                                              circle_size=circle_size,
                                              txt_shift=(x_shift, y_shift),
                                              font_size=font_size,
                                              body_part=self.body_part,
                                              threshold=self.threshold)
                for cnt, result in enumerate(pool.imap(constants, pose_data, chunksize=1)):
                    print(f"Section {result}/{len(pose_data)} complete...")
            pool.terminate()
            pool.join()
            concatenate_videos_in_folder(in_folder=self.temp_dir, save_path=self.video_save_path, gpu=False)
            video_timer.stop_timer()
            print(f"Video {self.video_save_path} complete (elapsed time: {video_timer.elapsed_time_str}s)...")

        self.timer.stop_timer()
        stdout_success(msg=f"{len(self.video_paths)} videos saved in {self.save_dir} directory", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, required=True, help='Path to directory containing DeepLabCut CSV files.')
    parser.add_argument('--video_dir', type=str, required=True, help='Directory containing video files corresponding with the names of the CSVs in data_dir.')
    parser.add_argument('--save_dir', type=str, required=True, help='Directory where to save the output videos.')
    parser.add_argument('--fps', type=float, required=True, help='Frames per second (int or float)')
    parser.add_argument('--body_part', type=str, required=True, help='Name of the body part to use for determining visibility (e.g., "snout", "center").')
    parser.add_argument('--threshold', type=float, required=True, help='Deeplabcut probability value. If below this value, animal is in dark box. If above, animal is in light box', default=0.1)
    parser.add_argument('--minimum_episode_duration', type=float, required=True, help='The shortest allowed visit to either the dark or the light box. Used to remove spurrious errors in the DLC tracking.', default=1)
    parser.add_argument('--core_cnt', type=int, required=True, help='The number of CPU cores to use. The higher the faster. -1 denotes all avaliable cores', default=-1)
    args = parser.parse_args()

    visualizer = LightDarkBoxPlotter(data_dir=args.data_dir,
                                    video_dir=args.video_dir,
                                    save_dir=args.save_dir,
                                    fps=args.fps,
                                    body_part=args.body_part,
                                    threshold=args.threshold,
                                    minimum_episode_duration=args.minimum_episode_duration,
                                    core_cnt=args.core_cnt)

    visualizer.run()














import os
import random
from copy import copy
from typing import Optional, Union

import cv2
import numpy as np

from simba.mixins.geometry_mixin import GeometryMixin
from simba.third_party_label_appenders.converters import create_yolo_yaml
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_int,
                                check_valid_boolean)
from simba.utils.enums import Options
from simba.utils.errors import FaultyTrainingSetError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (create_directory, get_fn_ext, read_img,
                                    read_json, recursive_file_search)


class COCOKeypoints2YoloSeg:

    def __init__(self,
                 coco_path: Union[str, os.PathLike],
                 img_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 train_size: float = 0.7,
                 verbose: bool = True,
                 greyscale: bool = False,
                 clahe: bool = False,
                 bbox_pad: Optional[int] = None):

        """
        Convert COCO keypoint annotations to YOLO segmentation format.

        .. image:: _static/img/cocokp_to_yolo_seg.webp
           :width: 600
           :align: center

        :param Union[str, os.PathLike] coco_path: Path to the input COCO JSON annotation file containing 'images', 'annotations', and 'categories'.
        :param Union[str, os.PathLike] img_dir: Directory where the input images referenced in the COCO file are stored.
        :param Union[str, os.PathLike] save_dir: Output directory where YOLO-formatted 'images' and 'labels' folders will be created.
        :param float train_size: Proportion of the dataset to use for training (between 0.1 and 0.99). Remaining goes to validation.
        :param bool verbose: Whether to print progress information during conversion.
        :param bool greyscale: If True, images will be loaded in greyscale.
        :param bool clahe: If True, Contrast Limited Adaptive Histogram Equalization (CLAHE) will be applied to images.
        :param Optional[int] bbox_pad: Optional padding to apply around the bounding box used to generate polygon segmentation.


        :example:
        >>>runner = COCOKeypoints2YoloSeg(coco_path=r"D:\cvat_annotations\frames\coco_keypoints_1\merged\merged_07032025.json",
        >>>                                img_dir=r"D:\cvat_annotations\frames",
        >>>                                save_dir=r"D:\cvat_annotations\yolo_07032025\bbox_seg",
        >>>                                clahe=False,
        >>>                                bbox_pad=None)
        >>>runner.run()
        """

        check_file_exist_and_readable(file_path=coco_path)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir')
        check_if_dir_exists(in_dir=img_dir, source=f'{self.__class__.__name__} img_dir')
        check_float(name=f'{self.__class__.__name__} train_size', value=train_size, max_value=0.99, min_value=0.1)
        if bbox_pad is not None: check_int(name=f'{self.__class__.__name__} bbox_pad', value=bbox_pad, min_value=1)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale', raise_error=True)
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe', raise_error=True)
        self.train_img_dir, self.val_img_dir = os.path.join(save_dir, 'images', 'train'), os.path.join(save_dir, 'images', 'val')
        self.train_lbl_dir, self.val_lbl_dir = os.path.join(save_dir, 'labels', 'train'), os.path.join(save_dir, 'labels', 'val')
        create_directory(paths=[self.train_img_dir, self.val_img_dir, self.train_lbl_dir, self.val_lbl_dir], overwrite=True)
        self.map_path = os.path.join(save_dir, 'map.yaml')
        self.coco_data = read_json(x=coco_path)
        check_if_keys_exist_in_dict(data=self.coco_data, key=['categories', 'images', 'annotations'], name=coco_path)
        self.map_dict = {i['id']: i['name'] for i in self.coco_data['categories']}
        map_ids = list(self.map_dict.keys())
        if sorted(map_ids) != list(range(len(map_ids))):
            self.map_id_lk = {}  # old: new
            new_map_dict = {}
            for cnt, v in enumerate(sorted(map_ids)):
                self.map_id_lk[v] = cnt
            for k, v in self.map_id_lk.items():
                new_map_dict[v] = self.map_dict[k]
            self.map_dict = copy(new_map_dict)
        else:
            self.map_id_lk = {k: k for k in self.map_dict.keys()}
        self.map_dict = {v: k for k, v in self.map_dict.items()}

        self.img_file_paths = recursive_file_search(directory=img_dir, extensions=Options.ALL_IMAGE_FORMAT_OPTIONS.value, case_sensitive=False, substrings=None, as_dict=True, raise_error=True)
        self.img_cnt = len(self.coco_data['images'])
        img_idx = list(range(len(self.coco_data['images']) + 1))
        self.train_idx = random.sample(img_idx, int(self.img_cnt * train_size))
        self.verbose, self.coco_path, self.img_dir, self.coco_path = verbose, coco_path, img_dir, coco_path
        self.save_dir, self.greyscale, self.clahe, self.bbox_pad = save_dir, greyscale, clahe, 1 if bbox_pad is None else bbox_pad

    def run(self):
        timer = SimbaTimer(start=True)
        for cnt in range(len(self.coco_data['images'])):
        #for cnt in range(500):
            img_data = self.coco_data['images'][cnt]
            check_if_keys_exist_in_dict(data=img_data, key=['width', 'height', 'file_name', 'id'], name=self.coco_path)
            _, img_name, ext = get_fn_ext(filepath=img_data['file_name'])
            if self.verbose:
                print(f'Processing annotation {cnt + 1}/{self.img_cnt} from COCO to YOLO ({img_name})...')
            if not img_name in self.img_file_paths.keys():
                raise NoFilesFoundError(msg=f'The file {img_name} could not be found in the {self.img_dir} directory', source=self.__class__.__name__)
            img = read_img(img_path=self.img_file_paths[img_name], greyscale=self.greyscale, clahe=self.clahe)
            img_h, img_w = img.shape[0], img.shape[1]
            if (img.shape[0] != img_data['height']) or (img.shape[1] != img_data['width']):
                raise FaultyTrainingSetError(msg=f'Image {img_name} is of shape {img.shape[0]}x{img.shape[1]}, but the coco data has been annotated on an image of {img_data["height"]}x{img_data["width"]}.')
            img_annotations = [x for x in self.coco_data['annotations'] if x['image_id'] == img_data['id']]
            roi_str = ' '
            if cnt in self.train_idx:
                label_save_path, img_save_path = os.path.join(self.train_lbl_dir, f'{img_name}.txt'), os.path.join(self.train_img_dir, f'{img_name}.png')
            else:
                label_save_path, img_save_path = os.path.join(self.val_lbl_dir, f'{img_name}.txt'), os.path.join(self.val_img_dir, f'{img_name}.png')
            for img_annotation in img_annotations:
                roi_str += '0 '
                check_if_keys_exist_in_dict(data=img_annotation, key=['bbox', 'keypoints', 'id', 'image_id', 'category_id'], name=str(self.coco_path))
                kps = np.array(img_annotation['keypoints']).reshape(-1, 3).astype(np.int32)
                bbox_kps = kps[kps[:, 2] != 0][:, 0:2]
                if bbox_kps.shape[0] < 2 or kps.shape[0] == 0:
                    continue
                bbox_kps = bbox_kps.reshape(-1, bbox_kps.shape[0], 2).astype(np.int32)
                seg_arr = GeometryMixin().bodyparts_to_polygon(data=bbox_kps, parallel_offset=self.bbox_pad)[0]
                seg_arr = np.array(seg_arr.exterior.coords).astype(np.int32)
                pts = np.unique(seg_arr, axis=0)
                center = np.mean(pts, axis=0)
                angles = np.arctan2(pts[:, 1] - center[1], pts[:, 0] - center[0])
                seg_arr = pts[np.argsort(angles)]
                seg_arr_x, seg_arr_y = np.clip(seg_arr[:, 0].flatten() / img_w, 0, 1), np.clip(seg_arr[:, 1].flatten() / img_h, 0, 1)
                kps = list(np.column_stack((seg_arr_x, seg_arr_y)).flatten())
                roi_str += ' '.join(str(x) for x in kps) + '\n'
            if roi_str:
                with open(label_save_path, mode='wt', encoding='utf-8') as f:
                    f.write(roi_str)
            cv2.imwrite(img_save_path, img)
        create_yolo_yaml(path=self.save_dir, train_path=self.train_img_dir, val_path=self.val_img_dir, names=self.map_dict, save_path=self.map_path)
        timer.stop_timer()
        if self.verbose: stdout_success(msg=f'Labelme to YOLO conversion complete. Data saved in directory {self.save_dir}.', elapsed_time=timer.elapsed_time_str)

# runner = COCOKeypoints2YoloSeg(coco_path=r"D:\cvat_annotations\frames\coco_keypoints_1\merged\merged_07032025.json",
#                                 img_dir=r"D:\cvat_annotations\frames",
#                                 save_dir=r"D:\cvat_annotations\yolo_07032025\bbox_seg",
#                                 clahe=False,
#                                 bbox_pad=None)
# runner.run()



__author__ = "Simon Nilsson"

import glob
import os
from typing import List, Union

import cv2
import imutils
import numpy as np

import simba
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_int, check_str,
                                check_valid_img_path, check_valid_lst)
from simba.utils.data import create_color_palettes
from simba.utils.enums import Paths, TextOptions
from simba.utils.errors import InvalidInputError
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    read_img)

WINDOW_NAME = "DEFINE POSE ESTIMATED BODY-PARTS"

class PoseConfigCreator(PlottingMixin):
    """

    Class for creating user-defined pose-estimation pipeline in SimBA through a GUI interface.

    ..seealso::
       `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/Pose_config.md>`__.

    :param str pose_name: Name of the user-defined pose-estimation setting.
    :param str no_animals: Number of animals in the user-defined pose-estimation setting.
    :param str img_path: Path to image representation of user-defined pose-estimation setting
    :param List[str] bp_list: Body-parts in the user-defined pose-estimation setting.
    :param List[int] animal_id_int_list: Integers representing the animal ID which the body-parts belong to.

    :examples:
    >>> pose_config_creator = PoseConfigCreator(pose_name="My_test_config", no_animals=2, img_path='simba/splash_050122.png', bp_list=['Ear', 'Nose', 'Left_ear', 'Ear', 'Nose', 'Left_ear'], animal_id_int_list= [1, 1, 1, 2, 2, 2])
    >>> pose_config_creator.launch()
    """

    def __init__(self,
                 pose_name: str,
                 animal_cnt: int,
                 img_path: Union[str, os.PathLike],
                 bp_list: List[str],
                 animal_id_int_list: List[int]):

        check_str(name="POSE CONFIG NAME", value=pose_name, allow_blank=False, raise_error=True)
        check_int(name="NUMBER OF ANIMALS", value=animal_cnt, min_value=1, raise_error=True)
        check_valid_img_path(path=img_path, raise_error=True)
        check_valid_lst(data=bp_list, source=f'{self.__class__.__name__} bp_list', valid_dtypes=(str,), min_len=1, raise_error=True)
        for bp_name in bp_list:
            if "," in bp_name: raise InvalidInputError(msg=f'Commas are not allowed in body-part names. A comma was found in body-part name: {bp_name}.', source=self.__class__.__name__)
        bp_list = [x.strip() for x in bp_list]
        if animal_cnt > 1:
            check_valid_lst(data=animal_id_int_list, source=f'{self.__class__.__name__} animal_id_int_list', valid_dtypes=(int,), min_len=len(bp_list), raise_error=True)
            if animal_cnt != len(list(set(animal_id_int_list))):
                raise InvalidInputError(msg=f'The number of animals (no_animals) is set to {animal_cnt}, but the number of unique IDs (animal_id_int_list) is set to {len(list(set(animal_id_int_list)))}.', source=self.__class__.__name__)
        self.pose_name, self.img_path, self.animal_cnt = pose_name, img_path, animal_cnt
        self.bp_list, self.animal_id_int_list = bp_list, animal_id_int_list

        PlottingMixin.__init__(self)
        self.img = read_img(img_path=img_path)
        self.img_h, self.img_w= int(self.img.shape[0]), int(self.img.shape[1])
        if self.img_w < 800:
            self.img = imutils.resize(self.img, width=800).astype(np.uint8)
            self.img_h, self.img_w = int(self.img.shape[0]), int(self.img.shape[1])

        self.circle_scale = self.get_optimal_circle_size(frame_size=(self.img_h, self.img_w), circle_frame_ratio=50)
        self.font_size, self.x_scale, self.y_scale = self.get_optimal_font_scales(text='Left click on body part XXXXXXXXXXWWW',accepted_px_width=self.img_w, accepted_px_height=int(self.img_h/10), text_thickness=4)
        cv2.namedWindow(WINDOW_NAME, cv2.WINDOW_NORMAL)
        self.overlay = self.img.copy()

        if self.animal_cnt > 1:
            for cnt, (bp_name, animal_number_id) in enumerate(zip(self.bp_list, self.animal_id_int_list)):
                self.bp_list[cnt] = f"{bp_name}_{animal_number_id}"
        self.color_lst = create_color_palettes(1, len(self.bp_list))[0]

    def launch(self):
        def draw_circle(event, x, y, flags, param):
            if event == cv2.EVENT_LBUTTONDOWN:
                cv2.circle(self.overlay, (x, int(y - self.side_img.shape[0])), self.circle_scale, self.color_lst[self.bp_cnt], -1)
                cv2.putText(self.overlay, str(self.bp_cnt + 1), (x + 4, int(y - self.side_img.shape[0])), cv2.FONT_HERSHEY_SIMPLEX, self.font_size, self.color_lst[self.bp_cnt], 4)
                self.cord_written = True

        for bp_cnt, bp_name in enumerate(self.bp_list):
            self.cord_written = False
            self.bp_cnt = bp_cnt
            self.side_img = np.zeros((int(self.img_h / 4), self.img_w, 3), np.uint8)
            if self.overlay.ndim != 3:
                self.side_img = cv2.cvtColor(self.side_img, cv2.COLOR_BGR2GRAY)
            cv2.putText(self.side_img, f"LEFT CLICK ON BODY-PART {bp_name}.", (10, 50), TextOptions.FONT.value, self.font_size, self.color_lst[bp_cnt], 4)
            img_concat = cv2.vconcat([self.side_img, self.overlay])
            cv2.namedWindow(WINDOW_NAME, cv2.WINDOW_NORMAL)
            cv2.imshow(WINDOW_NAME, img_concat)
            while not self.cord_written:
                cv2.setMouseCallback(WINDOW_NAME, draw_circle)
                img_concat = cv2.vconcat([self.side_img, self.overlay])
                cv2.namedWindow(WINDOW_NAME, cv2.WINDOW_NORMAL)
                cv2.imshow(WINDOW_NAME, img_concat)
                cv2.waitKey(1)
            cv2.destroyWindow(WINDOW_NAME)
        self.save()

    def save(self):
        overlay = cv2.resize(self.overlay, (250, 300))

        simba_cw = os.path.dirname(simba.__file__)
        img_dir = os.path.join(simba_cw, Paths.SCHEMATICS.value)
        check_if_dir_exists(in_dir=img_dir, source=self.__class__.__name__, create_if_not_exist=True)

        pose_name_path = os.path.join(simba_cw, Paths.PROJECT_POSE_CONFIG_NAMES.value)
        bp_path = os.path.join(simba_cw, Paths.SIMBA_BP_CONFIG_PATH.value)
        no_animals_path = os.path.join(simba_cw, Paths.SIMBA_NO_ANIMALS_PATH.value)
        for path in [pose_name_path, bp_path, no_animals_path]:
            check_file_exist_and_readable(file_path=path)
        prior_img_cnt = len(find_files_of_filetypes_in_directory(directory=img_dir, extensions=['.png'], raise_warning=False, raise_error=False))
        new_img_name = f"{prior_img_cnt + 1}.png"
        new_img_path = os.path.join(img_dir, new_img_name)
        joined_bp_lst = ",".join(self.bp_list)
        with open(pose_name_path, "a") as f:
            f.write(self.pose_name + "\n")
        with open(bp_path, "a") as fd:
            fd.write(joined_bp_lst + "\n")
        with open(no_animals_path, "a") as fd:
            fd.write(str(self.animal_cnt) + "\n")
        cv2.imwrite(new_img_path, overlay)


# pose_config_creator = PoseConfigCreator(pose_name="My_test_config", animal_cnt=2, img_path=r"C:\Users\sroni\OneDrive\Desktop\Screenshot 2024-09-22 151619.png", bp_list=['Ear', 'Nose', 'Left_ear', 'Ear', 'Nose', 'Left_ear'], animal_id_int_list= [1, 1, 1, 2, 2, 2])
# pose_config_creator.launch()
